from __future__ import annotations

from io import IOBase
from pydantic import BaseModel
from lzl.logging import logger
from lzl import load
from typing import Optional, Dict, Any, List, Union, Sequence, Callable, TYPE_CHECKING
from .types import SlackContext, SlackPayload
from .configs import SlackSettings

if load.TYPE_CHECKING:
    import slack_sdk
    import slack_sdk.web.async_client
    import slack_bolt
else:
    slack_sdk = load.LazyLoad("slack_sdk", install_missing = True)
    slack_bolt = load.LazyLoad("slack_bolt", install_missing = True)

if TYPE_CHECKING:
    from lzl.io import File
    from slack_sdk import WebClient
    from slack_sdk.web.async_client import AsyncWebClient
    from slack_sdk.models.attachments import Attachment
    from slack_sdk.models.blocks import Block

    from slack_bolt import App as SlackBoltApp, BoltContext, BoltRequest, BoltResponse
    from fastapi import APIRouter, FastAPI

logger.set_module_name(__name__, 'slack.client')

class SlackClient:

    verbose: Optional[bool] = False

    if TYPE_CHECKING:
        sapi: WebClient
        api: AsyncWebClient

    def __init__(
        self,
        token: Optional[str] = None,
        default_user: Optional[str] = None,
        default_channel: Optional[str] = None,
        username_mapping: Optional[Dict[str, Union[str, List[str]]]] = None,
        context_file: Optional['File'] = None,
        verbose: Optional[bool] = None,
        **kwargs
    ):
        """
        Slack Client
        """
        # from lazyops.imports._slacksdk import resolve_slack_sdk
        # resolve_slack_sdk(True)

        self.settings = SlackSettings()
        self.token = token or self.settings.bot_token
        self.disabled = self.token is None
        if self.token is None:
            logger.warning("No token provided. Disabling Slack client")
            return
        self._api: Optional['AsyncWebClient'] = None
        self._sapi: Optional['WebClient'] = None
        self._kwargs: Optional[Dict[str, Any]] = kwargs

        # Add Bolt handler
        self._bolt_app: Optional['SlackBoltApp'] = None

        if verbose is not None: self.verbose = verbose
        self.logger = logger
        self.default_user = default_user or self.settings.default_user
        self.default_user_id: Optional[str] = None

        self.default_channel = default_channel or self.settings.default_channel
        self.default_channel_id: Optional[str] = None
        self.ctx: SlackContext = self.settings.load_context(ctx_file = context_file)
        if username_mapping: self.set_username_mapping(username_mapping)

    def set_username_mapping(self, username_mapping: Dict[str, Union[str, List[str]]]):
        """
        Sets the username mapping
        """
        if self.ctx.username_mapping and username_mapping != self.ctx.username_mapping:
            from lzo.utils import update_dict
            self.ctx.username_mapping = update_dict(self.ctx.username_mapping, username_mapping)
            self.init(overwrite = True)
        else:
            self.ctx.username_mapping = username_mapping
            self.settings.save_context(self.ctx)


    @property
    def api(self) -> 'AsyncWebClient':
        """
        Returns the Async API
        """
        if self._api is None and not self.disabled:
            slack_sdk.__do_import__()
            from slack_sdk.web.async_client import AsyncWebClient
            kwargs = {k:v for k,v in self._kwargs.items() if k in AsyncWebClient.__init__.__code__.co_varnames and v is not None}
            self._api = AsyncWebClient(token = self.token, **kwargs)
        return self._api
    
    @property
    def sapi(self) -> 'WebClient':
        """
        Returns the Sync API
        """
        if self._sapi is None and not self.disabled:
            slack_sdk.__do_import__()
            from slack_sdk import WebClient
            kwargs = {k:v for k,v in self._kwargs.items() if k in WebClient.__init__.__code__.co_varnames and v is not None}
            self._sapi = WebClient(token = self.token, **kwargs)
            # self._sapi = WebClient(token = self.token, **self._kwargs)
        return self._sapi
    
    @property
    def bolt_app(self) -> 'SlackBoltApp':
        """
        Returns the Bolt App
        """
        if self._bolt_app is None and not self.disabled:
            slack_bolt.__do_import__()
            from slack_bolt import App
            kwargs = {k:v for k,v in self._kwargs.items() if k in App.__init__.__code__.co_varnames and v is not None}
            if 'signing_secret' not in kwargs and self.settings.signing_secret: kwargs['signing_secret'] = self.settings.signing_secret
            self._bolt_app = App(token = self.token, **kwargs)
        return self._bolt_app

    def get_users(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get users
        """
        resp = self.sapi.users_list(**kwargs)
        return resp["members"] if resp["ok"] else []

    async def aget_users(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get users
        """
        resp = await self.api.users_list(**kwargs)
        return resp["members"] if resp["ok"] else []

    def get_channels(self, types: Optional[List[str]] = None, include_users: Optional[bool] = True, **kwargs) -> List[Dict[str, Any]]:
        """
        Get channels
        """
        if not types:  types = ['public_channel','private_channel','im','mpim']
        if types and isinstance(types, list): types = ",".join(types)
        resp = self.sapi.conversations_list(types = types, **kwargs)
        channels =  resp["channels"] if resp["ok"] else []
        if include_users:
            resp = self.sapi.users_conversations(types = types, **kwargs)
            channels += resp["channels"] if resp["ok"] else []
        return channels
    
    async def aget_channels(self, types: Optional[List[str]] = None, include_users: Optional[bool] = True, **kwargs) -> List[Dict[str, Any]]:
        """
        Get channels
        """
        if not types:  types = ['public_channel','private_channel','im','mpim']
        if types and isinstance(types, list):
            types = ",".join(types)
        resp = await self.api.conversations_list(types = types, **kwargs)
        channels =  resp["channels"] if resp["ok"] else []
        if include_users:
            resp = await self.api.users_conversations(types = types, **kwargs)
            channels += resp["channels"] if resp["ok"] else []
        return channels

    @property
    def default_id(self) -> Optional[str]:
        """
        Default ID
        """
        if self.disabled: return
        return self.default_user_id or self.default_channel_id


    def configure_defaults(self):
        """
        Configure defaults
        """
        if self.default_user:
            self.default_user_id = self.ctx.lookup_id(self.default_user)
            if self.verbose: self.logger.info(f"Default user id: {self.default_user} -> {self.default_user_id}")
        if self.default_channel:
            self.default_channel_id = self.ctx.lookup_id(self.default_channel)
            if self.verbose: self.logger.info(f"Default channel id: {self.default_channel} -> {self.default_channel_id}")
        if self.default_id and self.verbose: self.logger.info(f"Default id: {self.default_id}")

    def init(self, overwrite: Optional[bool] = False):
        """
        Initialize
        """
        # Build Channels
        if self.disabled: return
        if not overwrite and self.ctx.initialized:
            self.configure_defaults()
            return

        users = self.get_users()
        for user in users:
            if aliases := self.ctx.username_mapping.get(user['name']):
                if isinstance(aliases, str): aliases = [aliases]
                for alias in aliases:
                    self.ctx.uids[user['id']] = alias
                    self.ctx.users[alias] =  user['id']
            else:
                self.ctx.uids[user['id']] = user['name']
            self.ctx.users[user['name']] =  user['id']


        channels = self.get_channels()
        for channel in channels:
            if not channel.get('name'): 
                self.ctx.channels[self.ctx.uids[channel['user']]] = channel['id']
                self.ctx.uids[channel['id']] = self.ctx.uids[channel['user']]
            else:
                self.ctx.channels[channel['name']] = channel['id']
                self.ctx.uids[channel['id']] = channel['name']
        
        self.ctx.initialized = True
        self.configure_defaults()
        self.settings.save_context(self.ctx)

    async def ainit(self, overwrite: Optional[bool] = False):
        """
        Initialize
        """
        # Build Channels
        if self.disabled: return
        if not overwrite and self.ctx.initialized:
            self.configure_defaults()
            return

        users = await self.aget_users()
        for user in users:
            if aliases := self.ctx.username_mapping.get(user['name']):
                if isinstance(aliases, str): aliases = [aliases]
                for alias in aliases:
                    self.ctx.uids[user['id']] = alias
                    self.ctx.users[alias] =  user['id']
            else:
                self.ctx.uids[user['id']] = user['name']
            self.ctx.users[user['name']] =  user['id']


        channels = await self.aget_channels()
        for channel in channels:
            if not channel.get('name'): 
                self.ctx.channels[self.ctx.uids[channel['user']]] = channel['id']
                self.ctx.uids[channel['id']] = self.ctx.uids[channel['user']]
            else:
                self.ctx.channels[channel['name']] = channel['id']
                self.ctx.uids[channel['id']] = channel['name']
        
        self.ctx.initialized = True
        self.configure_defaults()

    def lookup(self, user_or_channel: str) -> Optional[str]:
        """
        Return the ID of a user or channel
        """
        if self.disabled: return
        if user_or_channel in [
            self.default_id,
            self.default_channel_id,
        ]:
            return user_or_channel
        return self.ctx.lookup_id(user_or_channel) or user_or_channel

    def _validate(self) -> bool:
        """
        Validates the client
        """
        if self.disabled: return False
        if not self.ctx.initialized: self.init()
        return True
    
    async def _avalidate(self) -> bool:
        """
        Validates the client
        """
        if self.disabled: return False
        if not self.ctx.initialized: await self.ainit()
        return True

    def join_channel(self, channel_id: str, **kwargs):
        """
        Join a channel
        """
        if self.disabled: return
        if self.verbose: self.logger.info(f"Joining channel {channel_id}")
        self.sapi.conversations_join(channel = channel_id, **kwargs)

    async def ajoin_channel(self, channel_id: str, **kwargs):
        """
        Join a channel
        """
        if self.disabled: return
        if self.verbose: self.logger.info(f"Joining channel {channel_id}")
        await self.api.conversations_join(channel = channel_id, **kwargs)

    def message(
        self,
        message: Optional[str] = None,
        user_or_channel: Optional[str] = None,
        channel_id: Optional[str] = None,
        **kwargs
    ):
        """
        Send a message
        """
        if not self._validate(): return
        # if self.disabled: return
        if user_or_channel:
            channel_id = self.lookup(user_or_channel)
        elif channel_id is None:
            channel_id = self.default_id
        if not self.ctx.uids.get(channel_id): self.join_channel(channel_id)
        if channel_id is None:
            raise ValueError("No channel id provided")
        return self.sapi.chat_postMessage(channel = channel_id, text = message, **kwargs)
    

    async def amessage(
        self,
        message: str,
        user_or_channel: Optional[str] = None,
        channel_id: Optional[str] = None,
        **kwargs
    ):
        """
        Send a message
        """
        if not await self._avalidate(): return
        # if self.disabled: return
        if user_or_channel:
            channel_id = self.lookup(user_or_channel)
        elif channel_id is None:
            channel_id = self.default_id
        if not self.ctx.uids.get(channel_id): await self.ajoin_channel(channel_id)
        if channel_id is None:
            raise ValueError("No channel id provided")
        return await self.api.chat_postMessage(channel = channel_id, text = message, **kwargs)
    

    def upload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        if not self._validate(): return
        # if self.disabled: return
        channels = channels or [self.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: channels = [self.lookup(channel) for channel in channels]
        return self.sapi.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    async def aupload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        if not await self._avalidate(): return
        # if self.disabled: return
        channels = channels or [self.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: channels = [self.lookup(channel) for channel in channels]
        return await self.api.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    def temp_message(
        self,
        text: str,
        user: str,
        channel: Optional[str] = None,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        if not self._validate(): return
        # if self.disabled: return
        user = self.lookup(user)
        channel = self.lookup(channel) if channel else self.default_id
        return self.sapi.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = user, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )
    
    async def atemp_message(
        self,
        text: str,
        user: str,
        channel: Optional[str] = None,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        if not await self._avalidate(): return
        # if self.disabled: return
        user = self.lookup(user)
        channel = self.lookup(channel) if channel else self.default_id
        return await self.api.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = user, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )


    def __getitem__(self, name_or_id: str) -> 'SlackProxy':
        """
        Return the ID of a user or channel
        """
        if not self._validate(): return
        # if self.disabled: return
        return SlackProxy(name = name_or_id, uid = self.lookup(name_or_id) or name_or_id, client = self)

    def create_slash_command_endpoint(
        self,
        router: Union['APIRouter', 'FastAPI'],
        path: str,
        function: Callable,
        function_name: Optional[str] = None,
    ):
        
        function_name = function_name or function.__qualname__
        from fastapi.requests import Request

        async def slash_command(
            data: Request,
        ):
            payload = await data.form()
            return await function(SlackPayload(**payload))
        
        router.add_api_route(
            path = path,
            endpoint = slash_command,
            name = function_name,
            methods = ['POST'],
            include_in_schema=False,
        )
        return router


    # def create_interactive_endpoint(
    #     self,
    #     router: 'APIRouter',
    #     path: str,
    # ):
    #     """
    #     Create an interactive endpoint
    #     """
    #     from slack_sdk.web import WebClient
    #     from slack_bolt import App
    #     from slack_bolt.adapter.fastapi import SlackRequestHandler
    #     app = App(client = WebClient(token = self.token))
    #     handler = SlackRequestHandler(app = app)
    #     router.add_api_route(
    #         path = path,
    #         endpoint = handler,
    #         methods = ['POST'],
    #     )


class SlackProxy(BaseModel):

    name: str
    uid: str
    client: SlackClient

    model_config = {'arbitrary_types_allowed': True}


    def message(
        self,
        message: str,
        **kwargs
    ):
        """
        Send a message
        """
        return self.client.sapi.chat_postMessage(channel = self.uid, text = message, **kwargs)
    

    async def amessage(
        self,
        message: str,
        **kwargs
    ):
        """
        Send a message
        """
        return await self.client.api.chat_postMessage(channel = self.uid, text = message, **kwargs)
    
    def upload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        channels = channels or [self.client.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: 
            channels = [self.client.lookup(channel) for channel in channels]
        return self.client.sapi.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    

    async def aupload_file(
        self,
        file: Optional[Union[str, bytes, IOBase]] = None,
        content: Optional[str] = None,
        filename: Optional[str] = None,
        filetype: Optional[str] = None,
        initial_comment: Optional[str] = None,
        thread_ts: Optional[str] = None,
        title: Optional[str] = None,
        channels: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ):
        """
        Upload a file
        """
        channels = channels or [self.client.default_id]
        if not isinstance(channels, (list, tuple)): channels = [channels]
        if channels: 
            channels = [self.client.lookup(channel) for channel in channels]
        return await self.client.api.files_upload_v2(
            file = file, 
            content = content, 
            filename = filename, 
            filetype = filetype, 
            initial_comment = initial_comment, 
            thread_ts = thread_ts, 
            title = title, 
            channels = channels, 
            **kwargs
        )
    
    def temp_message(
        self,
        text: str,
        channel: str,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        channel = self.client.lookup(channel)
        return self.client.sapi.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = self.uid, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )

    async def atemp_message(
        self,
        text: str,
        channel: str,
        attachments: Optional[Sequence[Union[Dict, 'Attachment']]] = None,
        blocks: Optional[Sequence[Union[Dict, 'Block']]] = None,
        **kwargs,
    ):
        """
        Send a temporary message
        """
        channel = self.client.lookup(channel)
        return await self.client.api.chat_postEphemeral(
            text = text, 
            channel = channel, 
            user = self.uid, 
            attachments = attachments, 
            blocks = blocks,
            **kwargs
        )




from __future__ import annotations

"""
The Pathlib Implementation for Cloud Filesystems
"""

import io
import os
import hashlib
import datetime
import functools
import asyncio
import posixpath
import ntpath
import typing as t
from fsspec.callbacks import Callback
from aiopath.wrap import to_thread
from aiopath.handle import IterableAIOFile
from pydantic.types import ByteSize
from lzl.pool import ThreadPool
from lzl.io import aio
from lzl.types import PYDANTIC_VERSION, eproperty

from ..types.errors import FileExistsError
from ..types.misc import ObjectSize
from ..utils import logger
from ..path import PurePath, Path, FileMode, DEFAULT_ENCODING, ON_ERRORS, NEWLINE, Paths, PosixPath, WindowsPath
from ..path.flavours import _pathz_windows_flavour, _pathz_posix_flavour, _pathz_default_flavor
from .static.content_types import CONTENT_TYPE_BY_EXTENSION
# from .providers.main import AccessorLike, FileSystemLike, ProviderManager
from .utils import get_async_file, get_fsspec_file, AsyncFile



if t.TYPE_CHECKING:
    from .providers.main import AccessorLike
    from .main import FileLike, PathLike
    from fsspec.asyn import AsyncFileSystem
    from fsspec.spec import AbstractFileSystem
    from s3transfer.manager import TransferManager
    from ..configs.main import ProviderConfig


class CloudFileSystemPurePath(PurePath):
    _prefix: str = None
    _provider: str = None
    _win_pathz: t.ClassVar['PureCloudFileSystemWindowsPath'] = 'PureCloudFileSystemWindowsPath'
    _posix_pathz: t.ClassVar['PureCloudFileSystemPosixPath'] = 'PureCloudFileSystemPosixPath'

    def _init(self, template: t.Optional[PurePath] = None):
        from .providers.main import ProviderManager
        self._accessor = self._get_provider_accessor(self._prefix)

    def __new__(cls: t.Union[t.Type[CloudFileSystemPurePath], t.Type[PurePath]], *args):
        if cls is CloudFileSystemPurePath or issubclass(cls, CloudFileSystemPurePath):
            cls = cls._posix_pathz if os.name != 'nt' else cls._win_pathz
            cls = globals()[cls]
        return cls._from_parts(args)

    def _new(self, *parts):
        """Create a new `Path` child of same type."""
        return type(self)(*parts)
    
    @classmethod
    def _get_provider_accessor(cls, prefix: str) -> 'AccessorLike':
        """
        Returns the accessor for the provider
        """
        from .providers.main import ProviderManager
        return ProviderManager.get_accessor(prefix)
    
    

class PureCloudFileSystemPosixPath(CloudFileSystemPurePath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PureCloudFileSystemWindowsPath(CloudFileSystemPurePath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()



CloudPathT = t.TypeVar('CloudPathT', bound = 'CloudFileSystemPath')

class CloudFileSystemPath(Path, CloudFileSystemPurePath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_default_flavor
    _accessor: 'AccessorLike' = None
    _pathlike = posixpath
    _prefix = None
    _provider = None
    _win_pathz: t.ClassVar['CloudFileSystemWindowsPath'] = 'CloudFileSystemWindowsPath'
    _posix_pathz: t.ClassVar['CloudFileSystemPosixPath'] = 'CloudFileSystemPosixPath'

    def _init(self, template: t.Optional['CloudFileSystemPath'] = None):
        self._accessor = self._get_provider_accessor(self._prefix)
        self._closed = False
        self._fileio = None
        self._extra: t.Dict[str, t.Any] = {}

    def __new__(cls: t.Union[t.Type['CloudFileSystemPath'], t.Type[CloudPathT]], *parts, **kwargs) -> t.Union['CloudFileSystemPath', CloudPathT]:
        if cls is CloudFileSystemPath or issubclass(cls, CloudFileSystemPath):
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.string}")'

    def __str__(self):
        return self.string
    

    if PYDANTIC_VERSION == 2:
        if t.TYPE_CHECKING:
            from pydantic_core import core_schema
            from pydantic.annotated_handlers import GetCoreSchemaHandler, GetJsonSchemaHandler
            from pydantic.json_schema import JsonSchemaValue

        @classmethod
        def _get_filelike_(cls, *args, **kwargs) -> 'FileLike':
            """
            Returns the FileLike
            """
            from .main import get_filelike
            return get_filelike(*args, **kwargs)
        
        @classmethod
        def __get_pydantic_json_schema__(
            cls, 
            core_schema: 'core_schema.CoreSchema', 
            handler: 'GetJsonSchemaHandler'
        ) -> 'JsonSchemaValue':
            
            field_schema = handler(core_schema)
            field_schema.update(format = 'path', type = 'string')
            return field_schema

        @classmethod
        def __get_pydantic_core_schema__(
            cls, 
            source: type[t.Any], 
            handler: 'GetCoreSchemaHandler'
        ) -> 'core_schema.CoreSchema':
            """
            Get the Pydantic CoreSchema for the given source
            """
            from pydantic_core import core_schema, SchemaSerializer
            schema = core_schema.with_info_plain_validator_function(
                cls._validate,
                serialization = core_schema.to_string_ser_schema(),
            )
            cls.__pydantic_serializer__ = SchemaSerializer(schema)
            return schema

        @classmethod
        def _validate(cls, __input_value: t.Any, _: 'core_schema.ValidationInfo') -> 'FileLike':
            """
            Validator for Pydantic v2
            """
            return cls._get_filelike_(__input_value) if __input_value is not None else None

        def __hash__(self: 'FileLike') -> int:
            return hash(self.as_posix())

    @property
    def parent(self) -> 'FileLike':
        """
        The logical parent of the path.
        """
        return super().parent
    
    @property
    def filesys(self) -> t.Optional['AbstractFileSystem']:
        """
        Returns the filesystem object
        """
        return getattr(self._accessor, 'filesys', None)

    @property
    def afilesys(self) -> t.Optional['AsyncFileSystem']:
        """
        The async filesystem object associated with this path.
        """
        return getattr(self._accessor, 'afilesys', None)
    
    @property
    def fsconfig(self) -> 'ProviderConfig':
        """
        Returns the FileSystemConfig
        """
        return self._accessor.fsconfig
    

    @property
    def is_fsspec(self) -> bool:
        """
        Returns True if the path is a fsspec accessor
        """
        return self._accessor.is_fsspec
    
    @property
    def is_cloud_obj_(self) -> bool:
        """
        Returns True if the path is a cloud object
        """
        return self.is_fsspec

    @property
    def fspath_(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        # Mirrors _cloudpath
        if self._prefix in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        return self._pathlike.join(*self.parts)
    
    @property
    def fsuri_(self) -> t.Optional[str]:
        """
        Reconstructs the proper Fsspec URI
        """
        # Mirrors _cloudstr
        if self._prefix not in self.parts[0]:
            return f'{self._prefix}://' + '/'.join(self.parts)
        return f'{self._prefix}://' + '/'.join(self.parts[1:])
    
    @property
    def bucket_(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return self.parts[1] if self._prefix in self.parts[0] \
            else self.parts[0]
    
    @property
    def bucketstr_(self) -> t.Optional[str]:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        return f'{self._prefix}://{self.bucket_}'

    @property
    def string(self) -> str:
        """
        Return the string representation of the path with forward (/)
        """
        return self.fsuri_ if self.is_fsspec else self.posix_
    
    @property
    def posix_(self):
        """Return the string representation of the path with forward (/)
        slashes."""
        f = self._flavour
        return str(self).replace(f.sep, '/')
    
    @property
    def path_(self) -> str:
        """
        Returns the path as a string
        """
        return self.fsuri_ if self.is_fsspec else str(self)
    

    @property
    def pathkeys_(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        if self.bucket_ in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        if self.bucket_ in self.parts[1]: return self._pathlike.join(*self.parts[2:])
        return self._pathlike.join(*self.parts)

    def get_path_key(self, filename: t.Optional[str] = None) -> str:
        """
        Used to return relative/path/to/file.ext
        """
        filename = filename or self.name
        parts = None
        if self.bucket_ in self.parts[0]: parts = self.parts[1:-1]
        elif self.bucket_ in self.parts[1]: parts = self.parts[2:-1]
        else: parts = self.parts[:-1]
        return self._pathlike.join(*parts, filename)


    @property
    def filename_(self) -> str:
        """
        Returns the filename if is file, else ''
        """
        return self.parts[-1] if self.is_file() else ''

    @property
    def ext_(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def extension(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def stat_(self) -> os.stat_result:
        """
        Returns the stat results for path
        """
        return self.stat()

    @property
    def hash_(self) -> str:
        """
        Hash of file properties, to tell if it has changed
        """
        return self._accessor.ukey(self.fspath_)
    

    @property
    def info_(self):
        """
        Return info of path
        """
        return self._accessor.info(path=self.fspath_)

    @property
    def metadata_(self):
        """
        Return metadata of path
        """
        return self._accessor.metadata(self.fspath_)

    @property
    def path_info_(self):
        """
        Return info of path
        """
        return self._accessor.info(path=self.fspath_)
    

    @property
    def size_(self) -> t.Optional[t.Union[float, int]]:
        """
        Size in bytes of file
        """
        return self._accessor.size(self.fspath_) if self.is_file() else None

    @property
    def bytesize_(self) -> t.Optional['ByteSize']:
        """
        ByteSize in bytes of file
        """
        return self.bytesize()
    
    @property
    def fss3tm(self) -> t.Optional['TransferManager']:
        """
        Returns the s3t transfer manager
        """
        if self._accessor._s3t is not None:
            return self._accessor.get_s3t()
        raise NotImplementedError('S3 Transfer Manager is not Supported for this File System')
    
    
    @property
    def etag(self) -> t.Optional[str]:
        """
        Returns the etag of the file

        Cleans up the quoted string
        """
        _info = self.info()
        etag = _info.get('ETag', _info.get('etag'))
        return etag.strip('"') if etag else None
    
    @eproperty
    def checksum(self) -> str:
        """
        Returns the checksum of the file
        """
        from lzl.io.file.types.utils import generate_checksum
        return generate_checksum(self)
    
    @property
    def last_modified_(self) -> t.Optional[datetime.datetime]:
        """
        Returns the last modified time of the file
        """
        _info = self.info()
        return _info.get('LastModified', _info.get('last_modified'))
    
    @eproperty
    def content_type_(self) -> t.Optional[str]:
        """
        Returns the content type of the file
        """
        _info = self.info()
        return _info.get('ContentType', _info.get('content_type'))


    def open(self, mode: FileMode = 'r', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> t.IO[t.Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline)

    def aopen(self, mode: FileMode = 'r', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> AsyncFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        compression = infer doesn't work all that well.
        """
        return get_async_file(self._accessor.open(self.fspath_, mode=mode, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, buffering=buffering, **kwargs))

    def iter_raw(self, chunk_size: t.Optional[int] = None) -> t.Iterator[bytes]:
        """
        Iterates over the bytes of a file
        """
        from lzl.io.file.utils.decoders import ByteChunker
        chunker = ByteChunker(chunk_size = chunk_size)
        with self._accessor.open(self.fspath_, 'rb', block_size = chunk_size) as stream:
            for raw_stream_bytes in stream:
                yield from chunker.decode(raw_stream_bytes)
        yield from chunker.flush()

    def iter_text(self, chunk_size: int | None = None, encoding: str | None = None) -> t.Iterator[str]:
        """
        A str-iterator over the content
        """
        from lzl.io.file.utils.decoders import TextChunker, TextDecoder
        decoder = TextDecoder(encoding = encoding or "utf-8")
        chunker = TextChunker(chunk_size = chunk_size)
        for byte_content in self.iter_raw():
            text_content = decoder.decode(byte_content)
            yield from chunker.decode(text_content)

            text_content = decoder.flush()
            yield from chunker.decode(text_content)
            yield from chunker.flush()

    def iter_lines(self, chunk_size: int | None = None, encoding: str | None = None) -> t.Iterator[str]:
        """
        A line-by-line iterator over the file content.
        """
        from lzl.io.file.utils.decoders import LineDecoder
        decoder = LineDecoder()
        for text in self.iter_text(chunk_size = chunk_size, encoding = encoding):
            yield from decoder.decode(text)
        yield from decoder.flush()


    async def aiter_raw(self, chunk_size: t.Optional[int] = None) -> t.AsyncIterator[bytes]:
        """
        Iterates over the bytes of a file
        """
        from lzl.io.file.utils.decoders import ByteChunker
        chunker = ByteChunker(chunk_size = chunk_size)
        chunk_size = chunk_size if chunk_size is not None else -1
        async with await self.afilesys.open_async(self.fspath_, 'rb', block_size = chunk_size) as stream:
            raw_stream_bytes = await stream.read(chunk_size)
            while raw_stream_bytes:
                for chunk in chunker.decode(raw_stream_bytes):
                    yield chunk
                raw_stream_bytes = await stream.read(chunk_size)
        for chunk in chunker.flush():
            yield chunk

    async def aiter_text(self, chunk_size: int | None = None, encoding: str | None = None) -> t.AsyncIterator[str]:
        """
        A str-iterator over the content
        """
        from lzl.io.file.utils.decoders import TextChunker, TextDecoder
        decoder = TextDecoder(encoding = encoding or "utf-8")
        chunker = TextChunker(chunk_size = chunk_size)
        async for byte_content in self.aiter_raw():
            text_content = decoder.decode(byte_content)
            for chunk in chunker.decode(text_content):
                yield chunk
            text_content = decoder.flush()
            for chunk in chunker.decode(text_content):
                yield chunk
        
            for chunk in chunker.flush():
                yield chunk

    async def aiter_lines(self, chunk_size: int | None = None, encoding: str | None = None) -> t.AsyncIterator[str]:
        """
        A line-by-line iterator over the file content.
        """
        from lzl.io.file.utils.decoders import LineDecoder
        decoder = LineDecoder()
        async for text in self.aiter_text(chunk_size = chunk_size, encoding = encoding):
            for line in decoder.decode(text):
                yield line
        for line in decoder.flush():
            yield line


    def copy_to(self, dest: 'PathLike', overwrite: bool = False, chunk_size: t.Optional[int] = None, **kwargs) -> 'FileLike':
        """
        Copies this file to the destination path.
        """
        dst = self.get_pathlike_(dest)
        if not overwrite and dst.exists():
            raise FileExistsError(f"Destination file {dst} exists and overwrite is False")
        with dst.open('wb') as f:
            for chunk in self.iter_raw(chunk_size = chunk_size):
                f.write(chunk)
        return dst
    
    async def acopy_to(self, dest: 'PathLike', overwrite: bool = False, chunk_size: t.Optional[int] = None, **kwargs) -> 'FileLike':
        """
        Copies this file to the destination path.
        """
        dst = self.get_pathlike_(dest)
        if not overwrite and await dst.aexists():
            raise FileExistsError(f"Destination file {dst} exists and overwrite is False")
        async with dst.aopen('wb') as f:
            async for chunk in self.aiter_raw(chunk_size = chunk_size):
                await f.write(chunk)
        return dst

    def reader(self, mode: FileMode = 'r', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> t.IO[t.Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs)

    def areader(self, mode: FileMode = 'r', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> AsyncFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_async_file(self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs))

    def appender(self, mode: FileMode = 'a', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> t.IO[t.Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs)

    def aappender(self, mode: FileMode = 'a', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_async_file(self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs))

    def writer(self, mode: FileMode = 'w', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> t.IO[t.Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        #self.touch()
        return self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs)

    def awriter(self, mode: FileMode = 'w', buffering: int = -1, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, block_size: int = 5242880, compression: str = None, **kwargs: t.Any) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        #self.touch()
        return get_async_file(self._accessor.open(self.fspath_, mode=mode, buffering=buffering, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, **kwargs))

    def _read_chunked_with_tmgr(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None,  encoding: t.Optional[str] = DEFAULT_ENCODING, errors: str | None = ON_ERRORS,  **kwargs) -> t.Union[bytes, str]:
        # sourcery skip: class-extract-method
        """
        Read the data to the file with a transfer manager
        """
        # logger.info(f'Reading Chunked Data with TMGR: {self.fsuri_}')
        bytes_io = io.BytesIO()
        future = self.fss3tm.download(
            self.bucket_,
            self.get_path_key(self.name),
            bytes_io,
        )
        try:
            future.result()
        except Exception as e:
            if 'We encountered an internal error' in str(e):
                logger.error(f'Error reading chunked with TMGR: {e}')
                return self._read_chunked_default(mode = mode, start = start, end = end, encoding = encoding, errors = errors, **kwargs)
            raise e
        bytes_io.seek(0)
        if 'b' in mode: 
            if start: bytes_io.seek(start)
            if end: bytes_io.truncate(end)
            return bytes_io.getvalue()
        return io.TextIOWrapper(bytes_io, encoding = encoding, errors = errors, **kwargs).read()
    
    async def _aread_chunked_with_tmgr(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None,  encoding: t.Optional[str] = DEFAULT_ENCODING, errors: str | None = ON_ERRORS,  **kwargs) -> t.Union[bytes, str]:
        # sourcery skip: class-extract-method
        """
        Read the data to the file with a transfer manager
        """
        # logger.info(f'Reading Chunked Data with TMGR: {self.fsuri_}')
        bytes_io = io.BytesIO()
        future = self.fss3tm.download(
            self.bucket_,
            self.get_path_key(self.name),
            bytes_io,
        )
        try:
            await ThreadPool.asyncish(future.result)
        except Exception as e:
            if 'We encountered an internal error' in str(e):
                logger.error(f'Error reading chunked with TMGR: {e}')
                return await self._aread_chunked_default(mode = mode, start = start, end = end, encoding = encoding, errors = errors, **kwargs)
            raise e
        # future.result()
        bytes_io.seek(0)
        if 'b' in mode: 
            if start: bytes_io.seek(start)
            if end: bytes_io.truncate(end)
            return bytes_io.getvalue()
        return io.TextIOWrapper(bytes_io, encoding = encoding, errors = errors, **kwargs).read()
    
    
    def _read_chunked_default(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None,  encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> t.Union[bytes, str]:
        """
        Read the data to the file with default iterator
        """
        # logger.info(f'Reading Chunked Data with Default: {self.fsuri_}')
        _open_kwargs = {
            'encoding': encoding,
            'errors': errors,
            'newline': newline,
        } if 'b' not in mode else {}
        
        io_obj = io.BytesIO() if 'b' in mode else io.StringIO()
        _default = b'' if 'b' in mode else ''
        
        with self.open(mode = mode, **_open_kwargs) as f:
            for chunk in iter(lambda: f.read(self._accessor.fsconfig.read_chunking_size), _default):
                io_obj.write(chunk)
        io_obj.seek(0)
        if 'b' in mode: 
            if start: io_obj.seek(start)
            if end: io_obj.truncate(end)
        return io_obj.getvalue()
    
    async def _aread_chunked_default(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None,  encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> t.Union[bytes, str]:
        """
        Read the data to the file with default iterator
        """
        # logger.info(f'Reading Chunked Data with Default: {self.fsuri_}')
        _open_kwargs = {
            'encoding': encoding,
            'errors': errors,
            'newline': newline,
        } if 'b' not in mode else {}
        
        io_obj = aio.BytesIO() if 'b' in mode else aio.StringIO()
        async with self.aopen(mode = mode, **_open_kwargs) as f:
            block = await f.read(self._accessor.fsconfig.read_chunking_size)
            while block:
                io_obj.write(block)
                block = await f.read(self._accessor.fsconfig.read_chunking_size)
            # for chunk in await iter(lambda: f.read(self._accessor.fsconfig.read_chunking_size), _default):
            #     await io_obj.write(chunk)
        await io_obj.seek(0)
        if 'b' in mode: 
            if start: await io_obj.seek(start)
            if end: await io_obj.truncate(end)
        return io_obj.getvalue()

    def read_chunked_data(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, **kwargs) -> t.Union[int, ObjectSize]:
        # sourcery skip: extract-method
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        obj_size = self.get_object_size(self.size())
        if self._has_tmgr and (self.fsconfig.read_chunking_manager_default or obj_size > self.fsconfig.read_chunking_large_size):
            try:
                return self._read_chunked_with_tmgr(mode = mode, start = start, end = end, encoding = encoding, errors = errors, newline = newline, **kwargs)
            except Exception as e:
                logger.error(f'Error Reading Chunked Data: {e}')
        return self._read_chunked_default(mode = mode, start = start, end = end, encoding = encoding, errors = errors, newline = newline, **kwargs)

    def aread_chunked_data(self, mode: str = 'rb', start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE, **kwargs) -> t.Union[int, ObjectSize]:
        # sourcery skip: extract-method
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        obj_size = self.get_object_size(self.size())
        if self._has_tmgr and (self.fsconfig.read_chunking_manager_default or obj_size > self.fsconfig.read_chunking_large_size):
            try:
                return self._aread_chunked_with_tmgr(mode = mode, start = start, end = end, encoding = encoding, errors = errors, newline = newline, **kwargs)
            except Exception as e:
                logger.error(f'Error Reading Chunked Data: {e}')
        return self._aread_chunked_default(mode = mode, start = start, end = end, encoding = encoding, errors = errors, newline = newline, **kwargs)

    def read(self, mode: FileMode = 'rb', size: t.Optional[int] = -1, offset: t.Optional[int] = 0, **kwargs) -> t.Union[str, bytes]:
        """
        Read and return the file's contents.
        """
        with self.open(mode=mode, **kwargs) as file:
            if offset: file.seek(offset)
            return file.read(size)
        
    async def aread(self, mode: FileMode = 'rb', size: t.Optional[int] = -1, offset: t.Optional[int] = 0, **kwargs):
        """
        Read and return the file's contents.
        """
        async with self.aopen(mode=mode, **kwargs) as file:
            if offset: await file.seek(offset)
            return await file.read(size = size)

    def read_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS, **kwargs) -> str:
        """
        Read and return the file's contents.
        """
        if self.fsconfig.read_chunking_enabled:
            return self.read_chunked_data(mode = 'r', encoding = encoding, errors = errors, **kwargs)
        if hasattr(self.filesys, 'read_text'):
            return self.filesys.read_text(self.fspath_, encoding=encoding, errors=errors, **kwargs)
        with self.open('r', encoding=encoding, errors=errors) as file:
            return file.read()

    async def aread_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS, **kwargs) -> str:
        """
        Read and return the file's contents.
        """
        if self.fsconfig.read_chunking_enabled:
            return await self.aread_chunked_data(mode = 'r', encoding = encoding, errors = errors, **kwargs)
        async with self.aopen('r', encoding=encoding, errors=errors, **kwargs) as file:
            return await file.read()

    def read_bytes(self, start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None, **kwargs) -> bytes:
        """
        Read and return the file's contents.
        """
        if self.fsconfig.read_chunking_enabled:
            return self.read_chunked_data(mode = 'rb', start = start, end = end, **kwargs)
        if hasattr(self.filesys, 'read_bytes'):
            return self.filesys.read_bytes(self.fspath_, start=start, end=end)
        return self._accessor.cat_file(self.fspath_, start = start, end = end, **kwargs)

    async def aread_bytes(self, start: t.Optional[t.Any] = None, end: t.Optional[t.Any] = None, **kwargs) -> bytes:
        """
        Read and return the file's contents.
        """
        # async with self.aopen('rb') as file:
        #     return await file.read()
        if self.fsconfig.read_chunking_enabled:
            return await self.aread_chunked_data(mode = 'rb', start = start, end = end, **kwargs)
        return await self._accessor.acat_file(self.fspath_, start = start, end = end, **kwargs)

    def _guess_content_type(self) -> t.Optional[str]:
        """
        Guess the content type of the file
        """
        return CONTENT_TYPE_BY_EXTENSION.get(self.suffix, None)
    
    # def iter_bytes(self, chunk_size: int | None = None) -> t.Iterator[bytes]:
    #     """
    #     A byte-iterator over the decoded response content.
    #     """
    #     # _open_kwargs = {
    #     #     'encoding': encoding,
    #     #     'errors': errors,
    #     #     'newline': newline,
    #     # } if 'b' not in mode else {}
        
    #     # io_obj = io.BytesIO() if 'b' in mode else io.StringIO()
    #     # _default = b'' if 'b' in mode else ''
    #     chunk_size = chunk_size or self._accessor.fsconfig.read_chunking_size
    #     with self.open(mode = 'rb',) as f:
    #         for chunk in iter(lambda: f.read(chunk_size), b''):
    #             yield chunk

    # def iter_text(self, chunk_size: t.Optional[int] = None, **kwargs) -> t.Iterator[str]:
    #     """
    #     Iterates over the text in the file
    #     """


    @property
    def _has_tmgr(self) -> bool:
        """
        Checks if the file has a transfer manager
        """
        return self._accessor._s3t is not None
    
    def _write_chunked_with_tmgr(self, data: t.Union[bytes, str], obj_size: 'ObjectSize', encoding: t.Optional[str] = DEFAULT_ENCODING, **kwargs) -> 'ObjectSize':
        # sourcery skip: class-extract-method
        """
        Writes the data to the file with a transfer manager
        """
        if not isinstance(data, bytes): data = data.encode(encoding)
        obj_data = io.BytesIO(data)
        extra_args = {}
        if content_type := self._guess_content_type(): extra_args['ContentType'] = content_type
        future = self.fss3tm.upload(
            obj_data,
            self.bucket_,
            self.get_path_key(self.name),
            extra_args = extra_args,
        )
        # ThreadPool.threadpool_task(future.result)
        future.result()
        # Need to invalidate the cache since it 
        # isn't done in the fsspec library
        self.invalidate_cache()
        return obj_size

    async def _awrite_chunked_with_tmgr(self, data: t.Union[bytes, str], obj_size: 'ObjectSize', encoding: t.Optional[str] = DEFAULT_ENCODING, **kwargs) -> 'ObjectSize':
        # sourcery skip: class-extract-method
        """
        Writes the data to the file with a transfer manager
        """
        if not isinstance(data, bytes): data = data.encode(encoding)
        obj_data = io.BytesIO(data)
        extra_args = {}
        if content_type := self._guess_content_type(): extra_args['ContentType'] = content_type
        future = self.fss3tm.upload(
            obj_data,
            self.bucket_,
            self.get_path_key(self.name),
            extra_args = extra_args,
        )
        await ThreadPool.run_async(future.result)
        # Need to invalidate the cache since it 
        # isn't done in the fsspec library
        await self.ainvalidate_cache()
        return obj_size
    
    def _write_chunked_default(self, data: t.Union[bytes, str], obj_size: 'ObjectSize', encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> 'ObjectSize':
        """
        Writes the data to the file with default iterator
        """
        obj_data = io.BytesIO(data) if isinstance(data, bytes) else io.StringIO(data)
        mode = 'wb' if isinstance(data, bytes) else 'w'
        default_ = b'' if isinstance(data, bytes) else ''
        _open_kwargs = {
            'encoding': encoding,
            'errors': errors,
            'newline': newline,
        } if mode == 'w' else {}
        with self.open(mode = mode, **_open_kwargs) as f:
            for chunk in iter(lambda: obj_data.read(self._accessor.fsconfig.write_chunking_size), default_):
                f.write(chunk)
        if content_type := self._guess_content_type():
            self.setxattr(copy_kwargs={'ContentType': content_type})
        return obj_size
    
    async def _awrite_chunked_default(self, data: t.Union[bytes, str], obj_size: 'ObjectSize', encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> 'ObjectSize':
        """
        Writes the data to the file with default iterator
        """
        obj_data = aio.BytesIO(data) if isinstance(data, bytes) else aio.StringIO(data)
        mode = 'wb' if isinstance(data, bytes) else 'w'
        _open_kwargs = {
            'encoding': encoding,
            'errors': errors,
            'newline': newline,
        } if mode == 'w' else {}
        async with self.aopen(mode = mode, **_open_kwargs) as f:
            block = await obj_data.read(self._accessor.fsconfig.write_chunking_size)
            while block:
                await f.write(block)
                block = await obj_data.read(self._accessor.fsconfig.write_chunking_size)
        if content_type := self._guess_content_type():
            await self.asetxattr(copy_kwargs={'ContentType': content_type})
        return obj_size

    def write_chunked_data(self, data: t.Union[bytes, str], encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> t.Union[int, ObjectSize]:
        # sourcery skip: extract-method
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        obj_size = self.get_object_size(data)
        if self._has_tmgr and (self.fsconfig.write_chunking_manager_default or obj_size > self._accessor.fsconfig.write_chunking_large_size):
            return self._write_chunked_with_tmgr(data, obj_size)
        return self._write_chunked_default(data, obj_size, encoding = encoding, errors = errors, newline = newline)

    def awrite_chunked_data(self, data: t.Union[bytes, str], encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> t.Union[int, ObjectSize]:
        # sourcery skip: extract-method
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        obj_size = self.get_object_size(data)
        if self._has_tmgr and (self.fsconfig.write_chunking_manager_default or obj_size > self._accessor.fsconfig.write_chunking_large_size):
            return self._awrite_chunked_with_tmgr(data, obj_size)
        return self._awrite_chunked_default(data, obj_size, encoding = encoding, errors = errors, newline = newline)


    def write_bytes(self, data: bytes) -> t.Union[int, ObjectSize]:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        if self._accessor.fsconfig.write_chunking_enabled: 
            return self.write_chunked_data(data)
        view = memoryview(data)
        with self.open(mode='wb') as f:
            return f.write(data)

    async def awrite_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        if self.fsconfig.write_chunking_enabled: 
            return await self.awrite_chunked_data(data)
        view = memoryview(data)
        async with self.aopen(mode='wb') as f:
            return await f.write(data)

    def append_text(self, data: str, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        with self.open(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = f.write(data)
            n += f.write(newline)
            return n

    async def aappend_text(self, data: str, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.aopen(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = await f.write(data)
            n += await f.write(newline)
            return n

    def write_text(self, data: str, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if self._accessor.fsconfig.write_chunking_enabled: 
            return self.write_chunked_data(data, encoding = encoding, errors = errors, newline = newline)
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        # Check if the filesys has `write_text` method
        if hasattr(self.filesys, 'write_text'):
            return self.filesys.write_text(self.fspath_, data, encoding=encoding, errors=errors, newline=newline)
        with self.open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    async def awrite_text(self, data: str, encoding: t.Optional[str] = DEFAULT_ENCODING, errors: t.Optional[str] = ON_ERRORS, newline: t.Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if self.fsconfig.write_chunking_enabled: 
            return await self.awrite_chunked_data(data, encoding = encoding, errors = errors, newline = newline)
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.aopen(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return await f.write(data)


    def touch(self, truncate: bool = True, data = None, exist_ok: bool = True, **kwargs):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: self._accessor.stat(self.fspath_)
            # Avoid exception chaining
            except OSError: pass
            else: return
        try:
            self._accessor.touch(self.fspath_, truncate = truncate, data = data, **kwargs)
        except Exception as e:
            with self.open('wb') as f:
                f.write(b'')
                f.flush()


    async def atouch(self, truncate: bool = True, data = None, exist_ok: bool = True, **kwargs):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: await self._accessor.astat(self.fspath_)
            # Avoid exception chaining
            except OSError: pass
            else: return
        await self._accessor.atouch(self.fspath_, truncate = truncate, data = data, **kwargs)

    def mkdir(self, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """
        Create a new directory at this given path.
        """
        try: self._accessor.mkdir(self.fspath_, parents = parents, exist_ok = exist_ok)

        except FileNotFoundError:
            if not parents or self.parent == self: raise
            self.parent.mkdir(parents=True, exist_ok=True)
            self.mkdir(mode, parents=False, exist_ok=exist_ok)

        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not self.is_dir(): raise

    async def amkdir(self, parents: bool = True, exist_ok: bool = True, **kwargs):
        """
        Create a new directory at this given path.
        """
        try: await self._accessor.amkdir(self.fspath_, create_parents = parents, exist_ok = exist_ok, **kwargs)

        except FileNotFoundError:
            if not parents or self.parent == self: raise
            await self.parent.amkdir(parents=True, exist_ok=True, **kwargs)
            await self.amkdir(parents=False, exist_ok=exist_ok, **kwargs)

        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not await self.ais_dir(): raise

    async def amakedirs(self, parents: bool = True, exist_ok: bool = True):
        """
        Create a new directory at this given path.
        """
        try: await self._accessor.amakedirs(self.fspath_, exist_ok = exist_ok)

        except FileNotFoundError:
            if not parents or self.parent == self: raise
            await self.parent.amakedirs(exist_ok=True)
            await self.amakedirs(exist_ok=exist_ok)

        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not await self.ais_dir(): raise

    async def chmod(self, mode: int):
        """
        Change the permissions of the path, like os.chmod().
        """
        raise NotImplementedError

    async def achmod(self, mode: int):
        """
        Change the permissions of the path, like os.chmod().
        """
        raise NotImplementedError

    def lchmod(self, mode: int):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        raise NotImplementedError

    async def alchmod(self, mode: int):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        raise NotImplementedError

    def unlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: self._accessor.rm_file(self.fspath_)
        except FileNotFoundError:
            if not missing_ok: raise

    async def aunlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: await self._accessor.arm_file(self.fspath_)
        except FileNotFoundError:
            if not missing_ok: raise

    def rm(self, recursive: bool = False, maxdepth: int = None, missing_ok: bool = False):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try: self._accessor.rm(self.fspath_, recursive = recursive, maxdepth = maxdepth)
        except Exception as e:
            if not missing_ok: raise e


    async def arm(self, recursive: bool = False, maxdepth: int = None, missing_ok: bool = False):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try: await self._accessor.arm(self.fspath_, recursive = recursive, maxdepth = maxdepth)
        except Exception as e:
            if not missing_ok: raise e

    def rm_file(self, missing_ok: bool = True):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try:
            self._accessor.rm_file(self.fspath_)
            return True
        except Exception as e:
            if missing_ok: return False
            raise e from e


    async def arm_file(self, missing_ok: bool = True):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try:
            await self._accessor.arm_file(self.fspath_)
            return True
        except Exception as e:
            if missing_ok: return False
            raise e from e


    # async def aunlink(self, missing_ok: bool = False):
    #     """
    #     Remove this file or link.
    #     If the path is a directory, use rmdir() instead.
    #     """
    #     try: await self._accessor.aunlink(self.fspath_, missing_ok = missing_ok)
    #     except FileNotFoundError:
    #         if not missing_ok: raise

    def rmdir(self, force: bool = False, recursive: bool = True, skip_errors: bool = True):
        """
        Remove this directory.  The directory must be empty.
        """
        try:
            return self._accessor.rmdir(self.fspath_)
        except Exception as e:
            if force: return self._accessor.rmdir(self.fspath_, recursive = recursive)
            if skip_errors: return
            raise e


    async def armdir(self, force: bool = False, recursive: bool = True, skip_errors: bool = True):
        """
        Remove this directory.  The directory must be empty.
        """
        try:
            return await self._accessor.armdir(self.fspath_)
        except Exception as e:
            if force: return await self._accessor.armdir(self.fspath_, recursive = recursive)
            if skip_errors: return
            raise e

    def link_to(self, target: str):
        """
        Create a hard link pointing to a path named target.
        """
        raise NotImplementedError

    async def alink_to(self, target: str):
        """
        Create a hard link pointing to a path named target.
        """
        raise NotImplementedError

    def rename(self, target: t.Union[str, CloudPathT]) -> CloudPathT:
        """
        Rename this path to the target path.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        self._accessor.rename(self.fspath_, target)
        return type(self)(target)

    async def arename(self, target: t.Union[str, CloudPathT]) -> CloudPathT:
        """
        Rename this path to the target path.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        await self._accessor.arename(self.fspath_, target)
        return type(self)(target)

    def replace(self, target: str) -> CloudPathT:
        """
        Rename this path to the target path, overwriting if that path exists.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        self._accessor.replace(self.fspath_, target)
        return type(self)(target)

    async def areplace(self, target: str) -> CloudPathT:
        """
        Rename this path to the target path, overwriting if that path exists.
        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.
        Returns the new Path instance pointing to the target path.
        """
        await self._accessor.areplace(self.fspath_, target)
        return type(self)(target)

    def symlink_to(self, target: str, target_is_directory: bool = False):
        """
        Make this path a symlink pointing to the given path.
        Note the order of arguments (self, target) is the reverse of os.symlink's.
        """
        raise NotImplementedError

    async def asymlink_to(self, target: str, target_is_directory: bool = False):
        """
        Make this path a symlink pointing to the given path.
        Note the order of arguments (self, target) is the reverse of os.symlink's.
        """
        raise NotImplementedError

    def exists(self) -> bool:
        """
        Whether this path exists.
        """
        return self._accessor.exists(self.fspath_)


    async def aexists(self) -> bool:
        """
        Whether this path exists.
        """
        return await self._accessor.aexists(self.fspath_)

    if t.TYPE_CHECKING:
        def _make_child_relpath(self, part: t.Any) -> CloudPathT:
            ...

    @classmethod
    def cwd(cls: type) -> str:
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        cwd: str = os.getcwd()
        return cls(cwd)

    @classmethod
    def home(cls: type) -> CloudPathT:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        homedir: str = cls()._flavour.gethomedir(None)
        return cls(homedir)

    @classmethod
    async def ahome(cls: type) -> CloudPathT:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        coro = cls()._flavour.agethomedir(None)
        homedir: str = await coro
        return cls(homedir)

    def samefile(self, other_path: t.Union[CloudPathT, Paths]) -> bool:
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        if isinstance(other_path, Paths.__args__): other_path = CloudPathT(other_path)
        if isinstance(other_path, CloudPathT):
            try: other_st = other_path.stat()
            except AttributeError: other_st = self._accessor.stat(other_path)

        else:
            try: other_st = other_path.stat()
            except AttributeError: other_st = other_path._accessor.stat(other_path)
        return os.path.samestat(self.stat(), other_st)

    async def asamefile(self, other_path: t.Union[CloudPathT, Paths]) -> bool:
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        if isinstance(other_path, Paths.__args__): other_path = CloudPathT(other_path)
        if isinstance(other_path, CloudPathT):
            try: other_st = await other_path.astat()
            except AttributeError: other_st = await self._accessor.astat(other_path)

        else:
            try: other_st = await to_thread(other_path.stat)
            except AttributeError: other_st = await to_thread(other_path._accessor.stat, other_path)

        return os.path.samestat(await self.astat(),other_st)

    def listdir(self) -> t.List[CloudPathT]:
        """Return a list of the entries in the directory (as returned by
        os.listdir()).
        """
        # return self._accessor.listdir(self.fspath_)
        return [self._make_child_relpath(name) for name in self._accessor.listdir(self.fspath_)]

    async def alistdir(self) -> t.List[CloudPathT]:
        """Return a list of the entries in the directory (as returned by
        os.listdir()).
        """
        return [self._make_child_relpath(name) for name in await self._accessor.alistdir(self.fspath_)]
        # return [self._make_child_relpath(name) for name in await self._accessor.alistdir(self.fsuri_)]


    def iterdir(self) -> t.Iterable[CloudPathT]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for name in self._accessor.listdir(self.fspath_):
            if name in {'.', '..'}: continue
            yield self._make_child_relpath(name)

    async def aiterdir(self) -> t.AsyncIterable[CloudPathT]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        # for name in await self._accessor.alistdir(self):
        async for name in self._accessor.alistdir(self.fspath_):
            if name in {'.', '..'}: continue
            yield self._make_child_relpath(name)
    

    def walk(self) -> t.Iterable[t.Tuple[CloudPathT, t.List[CloudPathT], t.List[CloudPathT]]]:
        """Iterate over this subtree and yield a 3-tuple (dirpath, dirnames,
        filenames) for each directory in the subtree rooted at path
        (including path itself, if it is a directory).
        """
        top = self._make_child_relpath('.')
        dirs, nondirs = [], []
        for name in self._accessor.listdir(self.fspath_):
            if name in {'.', '..'}: continue
            (dirs if self._accessor.is_dir(self._make_child_relpath(name)) else nondirs).append(self._make_child_relpath(name))
        yield top, dirs, nondirs
        for name in dirs:
            new_path: CloudPathT = self._make_child_relpath(name)
            yield from new_path.walk()
        
    async def awalk(self) -> t.AsyncIterable[t.Tuple[CloudPathT, t.List[CloudPathT], t.List[CloudPathT]]]:
        """Iterate over this subtree and yield a 3-tuple (dirpath, dirnames,
        filenames) for each directory in the subtree rooted at path
        (including path itself, if it is a directory).
        """
        top = self._make_child_relpath('.')
        dirs, nondirs = [], []
        for name in await self._accessor.alistdir(self.fspath_):
            if name in {'.', '..'}: continue
            (dirs if await self._accessor.ais_dir(self._make_child_relpath(name)) else nondirs).append(self._make_child_relpath(name))
        yield top, dirs, nondirs
        for name in dirs:
            new_path: CloudPathT = self._make_child_relpath(name)
            async for path in new_path.awalk():
                yield path

    def glob(self, pattern: str = '*', as_path: bool = True) -> t.Iterable[t.Union[str, CloudPathT]]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        Warning: doesn't work as expected. Use Find Instead.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))
        #if self.is_fsspec:
        glob_pattern = self.fspath_ + ('/' if self.is_dir() and not self.fspath_.endswith('/') and not pattern.startswith('/') else '') +  pattern
        try:
            matches =  self._accessor.glob(glob_pattern)
            if not matches: return matches
            if self.is_fsspec: matches = [f'{self._prefix}://{m}' for m in matches]
            if as_path: matches = [type(self)(m) for m in matches]
            return matches
        except Exception as e:
            logger.error(e)
            return self.find(pattern = pattern, as_string = not as_path)

    async def aglob(self, pattern: str = '*', as_path: bool = True) -> t.AsyncIterable[CloudPathT]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        if not pattern: raise ValueError("Unacceptable pattern: {!r}".format(pattern))
        glob_pattern = self.fspath_ + ('/' if self.is_dir() and not self.fspath_.endswith('/') and not pattern.startswith('/') else '') +  pattern
        try:
            matches = await self._accessor.aglob(glob_pattern)
            if not matches: return matches
            if self.is_fsspec: matches = [f'{self._prefix}://{m}' for m in matches]
            if as_path: matches = [type(self)(m) for m in matches]
            return matches
        except Exception as e:
            logger.error(e)
            return await self.afind(pattern = pattern, as_string = not as_path)

    def find(self, pattern: str = "*",  as_string: bool = False, maxdepth: int = None, withdirs: bool = None, detail: bool = False) -> t.Union[t.List[str], t.List[CloudPathT]]:
        """
        List all files below path. Like posix find command without conditions
        """
        matches = self._accessor.find(path = self.fsuri_, maxdepth = maxdepth, withdirs = withdirs, detail = detail, prefix = pattern)
        if self.is_fsspec:
            matches = [f'{self._prefix}://{m}' for m in matches]
        if not as_string:
            matches = [type(self)(m) for m in matches]
        return matches

    async def afind(self, pattern: str = "*",  as_string: bool = False, maxdepth: int = None, withdirs: bool = None, detail: bool = False) -> t.Union[t.List[str], t.List[CloudPathT]]:
        """
        List all files below path. Like posix find command without conditions
        """
        matches = await self._accessor.afind(path = self.fsuri_, maxdepth = maxdepth, withdirs = withdirs, detail = detail, prefix = pattern)
        if self.is_fsspec:
            matches = [f'{self._prefix}://{m}' for m in matches]
        if not as_string:
            matches = [type(self)(m) for m in matches]
        return matches

    def rglob(self, pattern: str, as_path: bool = True) -> t.Iterable[t.Union[str, CloudPathT]]:
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        return self.glob(pattern = f'**/{pattern}', as_path = as_path)

    async def arglob(self, pattern: str) -> t.AsyncIterable[t.Union[str, CloudPathT]]:
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        return await self.aglob(f'**/{pattern}')

    def cat(self, recursive: bool = False, on_error: str = 'raise', **kwargs):
        """
        Fetch paths’ contents
        Parameters
        recursive: bool
            If True, assume the path(s) are directories, and get all the contained files

        on_error“raise”, “omit”, “return”
            If raise, an underlying exception will be raised (converted to KeyError if the type is in self.missing_exceptions);
            if omit, keys with exception will simply not be included in the output; if “return”, all keys are included in the output,
            but the value will be bytes or an exception instance.

        kwargs: passed to cat_file
        """
        return self._accessor.cat(self.fsuri_, recursive = recursive, on_error = on_error, **kwargs)

    async def acat(self, recursive: bool = False, on_error: str = 'raise', **kwargs):
        """
        Fetch paths’ contents
        Parameters
        recursive: bool
            If True, assume the path(s) are directories, and get all the contained files

        on_error“raise”, “omit”, “return”
            If raise, an underlying exception will be raised (converted to KeyError if the type is in self.missing_exceptions);
            if omit, keys with exception will simply not be included in the output; if “return”, all keys are included in the output,
            but the value will be bytes or an exception instance.

        kwargs: passed to cat_file
        """
        return await self._accessor.acat(self.fsuri_, recursive = recursive, on_error = on_error, **kwargs)

    def cat_file(self, as_bytes: bool = False, start: int = None, end: int = None, **kwargs):
        """
        Parameters
        start, end: int
            Bytes limits of the read. If negative, backwards from end, like usual python slices. Either can be None for start or end of file, respectively

        kwargs: passed to ``open()``.
        """
        res = self._accessor.cat_file(self.fsuri_, start = start, end = end, **kwargs)
        if not as_bytes and isinstance(res, bytes): res = res.decode('UTF-8')
        return res

    async def acat_file(self, as_bytes: bool = False, start: int = None, end: int = None, **kwargs):
        """
        Parameters
        start, end: int
            Bytes limits of the read. If negative, backwards from end, like usual python slices. Either can be None for start or end of file, respectively

        kwargs: passed to ``open()``.
        """
        res = await self._accessor.acat_file(self.fsuri_, start = start, end = end, **kwargs)
        if not as_bytes and isinstance(res, bytes): res = res.decode('UTF-8')
        return res

    def pipe(self, value: t.Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return self._accessor.pipe(self.fsuri_, value = value, **kwargs)

    async def apipe(self, value: t.Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return await self._accessor.apipe(self.fsuri_, value = value, **kwargs)

    def pipe_file(self, value: t.Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return self._accessor.pipe_file(self.fsuri_, value = value, **kwargs)

    async def apipe_file(self, value: t.Union[bytes, str], **kwargs):
        """
        Put value into path

        (counterpart to cat)
        """
        if not isinstance(value, bytes): value = value.encode('UTF-8')
        return await self._accessor.apipe_file(self.fsuri_, value = value, **kwargs)


    def absolute(self) -> CloudPathT:
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.
        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        raise NotImplementedError


    def resolve(self, strict: bool = False) -> CloudPathT:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        s: t.Optional[str] = self._flavour.resolve(self, strict=strict)

        if s is None:
            self.stat()
            path = self.absolute()
            s = str(path)

        # Now we have no symlinks in the path, it's safe to normalize it.
        normed: str = self._flavour.pathmod.normpath(s)
        obj = self._from_parts((normed,), init=False)
        obj._init(template=self)
        return obj

    async def aresolve(self, strict: bool = False) -> CloudPathT:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        s: t.Optional[str] = await self._flavour.aresolve(self, strict=strict)

        if s is None:
            await self.astat()
            path = await self.absolute()
            s = str(path)

        # Now we have no symlinks in the path, it's safe to normalize it.
        normed: str = self._flavour.pathmod.normpath(s)
        obj = self._from_parts((normed,), init=False)
        obj._init(template=self)
        return obj

    def stat(self) -> os.stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return self._accessor.stat(self.fspath_)

    async def astat(self) -> os.stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return await self._accessor.astat(self.fspath_)

    def info(self) -> t.Dict[str, t.Union[str, int, float, datetime.datetime, datetime.timedelta, t.List[str], t.Any]]:
        """
        Return the result of the info() system call on this path, like

        >>> {'ETag': '"etag"', 'LastModified': datetime.datetime(2025, 8, 14, 21, 10, 28, tzinfo=tzutc()), 'size': 78, 'name': 'path/in/bucket/key', 'type': 'file', 'StorageClass': 'STANDARD', 'VersionId': None, 'ContentType': 'binary/octet-stream'}        
        """
        return self._accessor.info(self.fspath_)

    async def ainfo(self) -> t.Dict[str, t.Union[str, int, float, datetime.datetime, datetime.timedelta, t.List[str], t.Any]]:
        """
        Return the result of the info() system call on this path, like
        os.stat() does.

        >>> {'ETag': '"etag"', 'LastModified': datetime.datetime(2025, 8, 14, 21, 10, 28, tzinfo=tzutc()), 'size': 78, 'name': 'path/in/bucket/key', 'type': 'file', 'StorageClass': 'STANDARD', 'VersionId': None, 'ContentType': 'binary/octet-stream'}
        """
        return await self._accessor.ainfo(self.fspath_)

    def metadata(self, refresh: t.Optional[bool] = False, **kwargs) -> t.Dict[str, t.Union[str, int, float, datetime.datetime, datetime.timedelta, t.List[str], t.Any]]:
        """
        Returns the metadata for the file
        """
        return self._accessor.metadata(self.fspath_, refresh = refresh, **kwargs)

    async def ametadata(self, refresh: t.Optional[bool] = False, **kwargs) -> t.Dict[str, t.Union[str, int, float, datetime.datetime, datetime.timedelta, t.List[str], t.Any]]:
        """
        Returns the metadata for the file
        """
        return await self._accessor.ametadata(self.fspath_, refresh = refresh, **kwargs)

    def size(self) -> int:
        """
        Return the size of the file, reported by os.path.getsize.
        """
        return self._accessor.size(self.fspath_)

    async def asize(self) -> int:
        """
        Return the size of the file, reported by os.path.getsize.
        """
        return await self._accessor.asize(self.fspath_)

    def bytesize(self) -> ByteSize:
        """
        Return the size of the file in bytes, reported by os.path.getsize().
        """
        return ByteSize(self.size())
    
    async def abytesize(self) -> ByteSize:
        """
        Return the size of the file in bytes, reported by os.path.getsize().
        """
        return ByteSize(await self.asize())

    def lstat(self) -> os.stat_result:
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        raise NotImplementedError

    async def alstat(self) -> os.stat_result:
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        raise NotImplementedError

    def owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        raise NotImplementedError

    async def aowner(self) -> str:
        """
        Return the login name of the file owner.
        """
        raise NotImplementedError

    def group(self) -> str:
        """
        Return the group name of the file gid.
        """
        raise NotImplementedError

    async def agroup(self) -> str:
        """
        Return the group name of the file gid.
        """
        raise NotImplementedError

    def is_dir(self) -> bool:
        """
        Whether this path is a directory.
        """
        return self._accessor.is_dir(self.fspath_)


    async def ais_dir(self) -> bool:
        """
        Whether this path is a directory.
        """
        return await self._accessor.ais_dir(self.fspath_)

    def is_symlink(self) -> bool:
        """
        Whether this path is a symbolic link.
        """
        raise NotImplementedError


    async def ais_symlink(self) -> bool:
        """
        Whether this path is a symbolic link.
        """
        raise NotImplementedError

    def is_file(self) -> bool:
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        return self._accessor.is_file(self.fspath_)

    async def ais_file(self) -> bool:
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        return await self._accessor.ais_file(self.fspath_)

    @staticmethod
    def get_pathlike_(path: 'PathLike') -> 'FileLike':
        """
        Returns the path of the file.
        """
        from lzl.io.file.spec.main import get_pathlike
        return get_pathlike(path)
    

    @classmethod
    def get_tempdir_(cls):
        """
        Returns the temporary directory for the current user.
        """
        import tempfile
        return cls.get_pathlike_(tempfile.gettempdir())

    def copy(self, dest: 'PathLike', recursive: bool = False, overwrite: bool = False, skip_errors: bool = False):
        """
        Copies the File to the Dir/File.
        """
        dest = self.get_pathlike_(dest)
        if dest.is_dir() and self.is_file():
            dest = dest.joinpath(self.filename_)
        if dest.exists() and not overwrite and dest.is_file():
            if skip_errors: return dest
            raise FileExistsError(f'File {dest.path_} exists')
        if dest.is_fsspec: self._accessor.copy(self.fspath_, dest.fspath_, recursive)
        else: self._accessor.get(self.fspath_, dest.string, recursive)
        return dest

    async def acopy(self, dest: 'PathLike', recursive: bool = False, overwrite: bool = False, skip_errors: bool = False):
        """
        Copies the File to the Dir/File.
        """
        dest = self.get_pathlike_(dest)
        if await dest.ais_dir() and await self.ais_file():
            dest = dest.joinpath(self.filename_)
        if await dest.aexists() and not overwrite and await dest.ais_file():
            if skip_errors: return dest
            raise FileExistsError(f'File {dest.path_} exists')
        if dest.is_fsspec: await self._accessor.acopy(self.fspath_, dest.fspath_, recursive = recursive)
        else: await self._accessor.aget(self.fspath_, dest.string, recursive = recursive)
        return dest

    def copy_file(self, dest: 'PathLike', recursive: bool = False, overwrite: bool = False, skip_errors: bool = False):
        """
        Copies this File to the the Dest Path
        """
        dest = self.get_pathlike_(dest)
        if dest.is_dir() and self.is_file():
            dest = dest.joinpath(self.filename_)
        if dest.exists() and not overwrite and dest.is_file():
            if skip_errors: return dest
            raise FileExistsError(f'File {dest.path_} exists')
        if dest.is_fsspec: self._accessor.copy(self.fspath_, dest.fspath_, recursive)
        else: self._accessor.get(self.fspath_, dest.path_, recursive)
        return dest

    async def acopy_file(self, dest: 'PathLike', recursive: bool = False, overwrite: bool = False, skip_errors: bool = False):
        """
        Copies this File to the the Dest Path
        """
        dest = self.get_pathlike_(dest)
        if await dest.ais_dir() and await self.ais_file():
            dest = dest.joinpath(self.filename_)
        if await dest.aexists() and not overwrite and await dest.ais_file():
            if skip_errors: return dest
            raise FileExistsError(f'File {dest.path_} exists')
        if dest.is_fsspec: await self._accessor.acopy(self.fspath_, dest.fspath_, recursive = recursive)
        else: await self._accessor.aget(self.fspath_, dest.string, recursive = recursive)
        return dest
    

    def localize(self, filename: t.Optional[str] = None, output_dir: t.Optional[Paths] = None, unique: t.Optional[bool] = False, cleanup_on_exit: t.Optional[bool] = False, **kwargs) -> 'FileLike':
        """
        Localize the file

        Otherwise, it will copy the file to the specified output directory or temporary directory.
        """
        if not self.is_fsspec: return self
        if output_dir is None: output_dir = self.get_tempdir_()
        else: output_dir = self.get_pathlike_(output_dir)
        if not filename:
            if unique:
                from lzo.utils import create_unique_id
                filename = f'{create_unique_id()}{self.suffix}'
            else:
                filename = self.name
        dest = output_dir.joinpath(filename)
        if dest.is_fsspec: self._accessor.copy(self.fspath_, dest.fspath_)
        else: self._accessor.get(self.fspath_, dest.path_)
        if cleanup_on_exit:
            from lzl.io.registry import add_file_to_cleanup
            add_file_to_cleanup(dest)
        return dest
    

    async def alocalize(self, filename: t.Optional[str] = None, output_dir: t.Optional[Paths] = None, unique: t.Optional[bool] = False, cleanup_on_exit: t.Optional[bool] = False, **kwargs) -> 'FileLike':
        """
        Localize the file - If this is already a local file (not cloud - it doesn't do anything)

        Otherwise, it will copy the file to the specified output directory or temporary directory.
        """
        if not self.is_fsspec: return self
        if output_dir is None: output_dir = self.get_tempdir_()
        else: output_dir = self.get_pathlike_(output_dir)
        if not filename:
            if unique:
                from lzo.utils import create_unique_id
                filename = f'{create_unique_id()}{self.suffix}'
            else: filename = self.name
        dest = output_dir.joinpath(filename)
        if dest.is_fsspec: await self._accessor.acopy(self.fspath_, dest.fspath_)
        else: await self._accessor.aget(self.fspath_, dest.string)
        if cleanup_on_exit:
            from lzl.io.registry import add_file_to_cleanup
            add_file_to_cleanup(dest)
        return dest
    

    def bytesio(self, **kwargs) -> io.BytesIO:
        """
        Convert the file to a BytesIO object.
        """
        bytes_obj = io.BytesIO()
        bytes_obj.write(self.read_bytes())
        bytes_obj.seek(0)
        return bytes_obj
    
    async def abytesio(self, **kwargs) -> io.BytesIO:
        """
        Convert the file to a BytesIO object.
        """
        bytes_obj = io.BytesIO()
        bytes_obj.write(await self.aread_bytes())
        bytes_obj.seek(0)
        return bytes_obj

    def textio(self, **kwargs) -> io.StringIO:
        """
        Convert the file to a StringIO object.
        """
        string_obj = io.StringIO()
        string_obj.write(self.read_text())
        string_obj.seek(0)
        return string_obj

    async def atextio(self, **kwargs) -> io.StringIO:
        """
        Convert the file to a StringIO object.
        """
        string_obj = io.StringIO()
        string_obj.write(await self.aread_text())
        string_obj.seek(0)
        return string_obj
    

    """
    Serialization Helper for file-like objects.
    """
    def yaml(self, **kwargs) -> t.Union[t.Dict[str, t.Any], t.List[t.Any]]:
        """
        Parses the file contents as yaml
        """
        import yaml
        return yaml.safe_load(self.read_text( **kwargs))
    
    async def ayaml(self, **kwargs) -> t.Union[t.Dict[str, t.Any], t.List[t.Any]]:
        """
        Parses the file contents as yaml
        """
        import yaml
        return yaml.safe_load(await self.aread_text( **kwargs))

    def json(self, **kwargs) -> t.Union[t.Dict[str, t.Any], t.List[t.Any]]:
        """
        Parses the file contents as json
        """
        from lzl.io.registry import json_ser
        try:
            return json_ser.loads(self.read_text( **kwargs))
        except Exception as e:
            from lzl.logging import logger
            logger.trace(f"Failed to parse JSON from {self.path_}", e)
            import json
            return json.loads(self.read_text( **kwargs))
        
    
    async def ajson(self, **kwargs) -> t.Union[t.Dict[str, t.Any], t.List[t.Any]]:
        """
        Parses the file contents as json
        """
        from lzl.io.registry import json_ser
        try:
            return await json_ser.aloads(await self.aread_text( **kwargs))
        except Exception as e:
            from lzl.logging import logger
            logger.trace(f"Failed to parse JSON from {self.path_}", e)
            import json
            return json.loads(await self.aread_text( **kwargs))

    @classmethod
    def register_loader(cls, ext: str, loader: t.Union[t.Callable[['FileLike'], None], t.Awaitable['FileLike', None]], overwrite: t.Optional[bool] = None) -> None:
        """
        Registers a file loader for a specific file extension.
        """
        from lzl.io.file.registry import register_loader
        register_loader(ext, loader, overwrite)

    def loads(self, *args, **kwargs) -> t.Any:
        """
        Automatically Load this File using the registered loaders based on the extension.

        If no loader is found, an error is raised.

        Use `File.register_loader` to register a new loader.
        """
        from lzl.io.file.registry import get_file_loader
        loader_func = get_file_loader(self.suffix)
        if ThreadPool.is_coro(loader_func):
            return ThreadPool.run_sync(loader_func, self, *args, **kwargs)
        return loader_func(self, *args, **kwargs)
    
    async def aloads(self, *args, **kwargs) -> t.Any:
        """
        Automatically Load this File using the registered loaders based on the extension.

        If no loader is found, an error is raised.

        Use `File.register_loader` to register a new loader.
        """
        from lzl.io.file.registry import get_file_loader
        loader_func = get_file_loader(self.suffix)
        if ThreadPool.is_coro(loader_func):
            return await loader_func(self, *args, **kwargs)
        return ThreadPool.arun(loader_func, self, *args, **kwargs)

    """
    End Serialization Helpers
    """

    def _download_with_tmgr(
        self, 
        dest: 'FileLike',
        callbacks: t.Optional[t.Dict[str, t.Callable]] = None,
        **kwargs
    ) -> 'FileLike':
        """
        Downloads the current file to the dest
        """
        future = self.fss3tm.download(
            self.bucket_,
            self.get_path_key(self.name),
            dest.as_posix(),
            subscribers = callbacks,
            **kwargs,
        )
        future.result()

        return dest

    def download(
        self, 
        dest: 'PathLike',
        filename: t.Optional[str] = None,
        overwrite: t.Optional[bool] = None,
        callbacks: t.Optional[t.Dict[str, t.Callable]] = None,
        **kwargs
    ) -> 'FileLike':
        """
        Downloads the dest to the current path
        """
        dest = self.get_pathlike_(dest)
        new = dest.joinpath(filename or self.name) if self.is_dir() else self
        if new.exists() and not overwrite:
            raise FileExistsError(f'File {new.path_} exists and overwrite is False')
        if self._has_tmgr:
            return self._download_with_tmgr(dest, callbacks, **kwargs)
        return self.copy_file(dest, overwrite = overwrite, **kwargs)
    
    def _upload_with_tmgr(
        self, 
        src: 'FileLike',
        callbacks: t.Optional[t.Dict[str, t.Callable]] = None,
        **kwargs
    ) -> 'FileLike':
        """
        Uploads the src to the current path
        """
        future = self.fss3tm.upload(
            src.as_posix(),
            self.bucket_,
            self.get_path_key(self.name),
            subscribers = callbacks,
            **kwargs,
        )
        future.result()
        self.invalidate_cache()
        return src

    def upload(
        self, 
        src: 'PathLike',
        filename: t.Optional[str] = None, 
        overwrite: t.Optional[bool] = None,
        callbacks: t.Optional[t.Dict[str, t.Callable]] = None,
        **kwargs
    ) -> 'FileLike':
        """
        Uploads the src to the current path
        """
        src = self.get_pathlike_(src)
        new = self.joinpath(filename or src.name) if self.is_dir() else self
        if new.exists() and not overwrite:
            raise FileExistsError(f'File {new.path_} exists and overwrite is False')
        if not src.exists():
            raise FileNotFoundError(f'File {src.path_} does not exist')
        if self._has_tmgr:
            return self._upload_with_tmgr(src, callbacks, **kwargs)
        return self.copy_file(src, overwrite = overwrite, **kwargs)
        

    def put(self, src: 'PathLike', recursive: bool = False, callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy file(s) from src to this FilePath
        WIP support for cloud-to-cloud
        """
        src = self.get_pathlike_(src)
        assert not src.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return self._accessor.put(src.string, self.fspath_, recursive=recursive, callback=callback, **kwargs)

    async def aput(self, src: 'PathLike', recursive: bool = False, callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy file(s) from src to this FilePath
        WIP support for cloud-to-cloud
        """
        src = self.get_pathlike_(src)
        assert not src.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return await self._accessor.aput(src.string, self.fspath_, recursive=recursive, callback=callback, **kwargs)

    def put_file(self, src: 'PathLike', callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy single file to remote
        WIP support for cloud-to-cloud
        """
        src = self.get_pathlike_(src)
        assert not src.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return self._accessor.put_file(src.string, self.fspath_, callback=callback, **kwargs)

    async def aput_file(self, src: 'PathLike', callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy single file to remote
        WIP support for cloud-to-cloud
        """
        src = self.get_pathlike_(src)
        assert not src.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return await self._accessor.aput_file(src.string, self.fspath_, callback=callback, **kwargs)

    def get(self, dest: 'PathLike', recursive: bool = False, callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy the remote file(s) to dest (local)
        WIP support for cloud-to-cloud
        """
        dest = self.get_pathlike_(dest)
        assert not dest.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return self._accessor.get(self.fspath_, dest.string, recursive=recursive, callback=callback, **kwargs)

    async def aget(self, dest: 'PathLike', recursive: bool = False, callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copy the remote file(s) to dest (local)
        WIP support for cloud-to-cloud
        """
        dest = self.get_pathlike_(dest)
        assert not dest.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return await self._accessor.aget(self.fspath_, dest.string, recursive=recursive, callback=callback, **kwargs)

    def get_file(self, dest: 'PathLike', callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copies this file to dest (local)
        WIP support for cloud-to-cloud
        """
        dest = self.get_pathlike_(dest)
        assert not dest.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return self._accessor.get_file(self.fspath_, dest.string, callback=callback, **kwargs)

    async def aget_file(self, dest: 'PathLike', callback: t.Optional[t.Callable] = Callback(), **kwargs):
        """
        Copies this file to dest (local)
        WIP support for cloud-to-cloud
        """
        dest = self.get_pathlike_(dest)
        assert not dest.is_fsspec, 'Cloud to Cloud support not supported at this time'
        return await self._accessor.aget_file(self.fspath_, dest.string, callback=callback, **kwargs)


    def is_mount(self) -> bool:
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        return False if not self.exists() or not self.is_dir() else False
        #raise NotImplementedError


    async def ais_mount(self) -> bool:
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        return False if not await self.aexists() or not await self.ais_dir() else False
        #raise NotImplementedError


    def is_block_device(self) -> bool:
        """
        Whether this path is a block device.
        """
        return False
        #raise NotImplementedError

    async def ais_block_device(self) -> bool:
        """
        Whether this path is a block device.
        """
        return False
        #raise NotImplementedError

    def is_char_device(self) -> bool:
        """
        Whether this path is a character device.
        """
        return False
        #raise NotImplementedError


    async def ais_char_device(self) -> bool:
        """
        Whether this path is a character device.
        """
        return False

    def is_fifo(self) -> bool:
        """
        Whether this path is a FIFO.
        """
        return False


    async def ais_fifo(self) -> bool:
        """
        Whether this path is a FIFO.
        """
        return False


    def is_socket(self) -> bool:
        """
        Whether this path is a socket.
        """
        return False


    async def ais_socket(self) -> bool:
        """
        Whether this path is a socket.
        """
        return False


    def expanduser(self) -> CloudPathT:
        """ Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        if (not self._drv and not self._root and self._parts and self._parts[0][:1] == '~'):
            homedir = self._flavour.gethomedir(self._parts[0][1:])
            return self._from_parts([homedir] + self._parts[1:])
        return self

    async def aexpanduser(self) -> CloudPathT:
        """ Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        if (not self._drv and not self._root and self._parts and self._parts[0][:1] == '~'):
            homedir = await self._flavour.agethomedir(self._parts[0][1:])
            return self._from_parts([homedir] + self._parts[1:])
        return self

    def iterdir(self) -> t.Iterable[CloudPathT]:
        names = self._accessor.listdir(self)
        for name in names:
            if name in {'.', '..'}: continue
        yield self._make_child_relpath(name)

    async def aiterdir(self) -> t.AsyncIterable[CloudPathT]:
        names = await self._accessor.alistdir(self)
        for name in names:
            if name in {'.', '..'}: continue
        yield self._make_child_relpath(name)

    async def als(
        self,
        recursive: bool = False,
        detail: t.Optional[bool] = False,
        versions: t.Optional[bool] = False,
        refresh: t.Optional[bool] = False,
        as_path: t.Optional[bool] = True,
        files_only: t.Optional[bool] = True,
        prettify: t.Optional[bool] = True,
        **kwargs,
    ) -> t.Union[t.List[CloudPathT], t.List[t.Dict[str, t.Any]]]:
        """
        Return a list of the files in this directory.
        """
        
        ls_partial = functools.partial(self.afilesys._ls, detail=True, versions=versions, refresh=refresh)

        async def _ls(path: t.Dict[str, t.Any]):
            """
            Inner function to handle recursive ls
            """
            ps = []
            if path.get('type', path.get('StorageClass', '')).lower() == 'directory' \
                and recursive:
                new_paths = await ls_partial(path['Key'])
                results = await asyncio.gather(*[_ls(p) for p in new_paths])
                for result in results:
                    ps.extend(result)
                if not files_only: ps.append(path)
            if files_only and path.get('type', path.get('StorageClass', '')).lower() != 'directory':
                ps.append(path)
            return ps

        from pydantic.types import ByteSize
        from lzl.pool import ThreadPool
        # from lzl.pool import async_map
        # from fileio.utils.pooler import async_map
        paths: t.List[t.Dict[str, t.Any]] = await ls_partial(self.fspath_)
        all_paths = []
        async for path in ThreadPool.aiterate(_ls, paths):
            all_paths.extend(path)

        if detail: 
            for p in all_paths:
                p['Key'] = f'{self._prefix}://{p["Key"]}'
                if prettify: p['SizePretty'] = ByteSize(p['Size']).human_readable()
            if as_path:
                p['File'] = type(self)(p['Key'])
            return all_paths
        final_paths = []
        for path in all_paths:
            if as_path: 
                path_str = f'{self._prefix}://{path["Key"]}'
                path = type(self)(path_str)
            final_paths.append(path)
        return final_paths
        

    def _raise_closed(self):
        raise ValueError("I/O operation on closed path")

    def _raise_open(self):
        raise ValueError("I/O operation on already open path")

    # We sort of assume that it will be used to open a file
    def __enter__(self):
        #if self._fileio: self._raise_open()
        #if not self._fileio:
        #    self._fileio = self.open()
        if self._closed: self._raise_closed()
        return self

    def __exit__(self, t, v, tb):
        self._closed = True

    async def __aenter__(self):
        if self._closed: self._raise_closed()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._closed = True

    """
    Other Methods
    """
    @t.overload
    def url(self, expires: int = 3600, client_method: str = 'get_object', **kwargs) -> str:
        """
        Returns a presigned URL for this file object
        """
        ...
        
    def url(self, **kwargs):
        return self._accessor.url(self.fspath_, **kwargs)

    @t.overload
    async def aurl(self, expires: int = 3600, client_method: str = 'get_object', **kwargs) -> str:
        """
        Returns a presigned URL for this file object
        """
        ...

    async def aurl(self, **kwargs):
        return await self._accessor.aurl(self.fspath_, **kwargs)

    @t.overload
    def setxattr(self, copy_kwargs: t.Optional[t.Dict[str, t.Any]] = None, **kwargs) -> None:
        """
        Set the extended attributes for this file

        Set metadata.

        Attributes have to be of the form documented in the
        `Metadata Reference`_.

        Parameters
        ----------
        kwargs : key-value pairs like field="value", where the values must be
            strings. Does not alter existing fields, unless
            the field appears here - if the value is None, delete the
            field.
        copy_kwargs : dict, optional
            dictionary of additional params to use for the underlying
            s3.copy_object.

        Examples
        --------
        >>> mys3file.setxattr(attribute_1='value1', attribute_2='value2')  # doctest: +SKIP
        # Example for use with copy_args
        >>> mys3file.setxattr(copy_kwargs={'ContentType': 'application/pdf'},
        ...     attribute_1='value1')  # doctest: +SKIP

        """
        ...

    def setxattr(self, **kwargs):
        return self._accessor.setxattr(self.fspath_, **kwargs)

    @t.overload
    async def asetxattr(self, copy_kwargs: t.Optional[t.Dict[str, t.Any]] = None, **kwargs) -> None:
        """
        Set the extended attributes for this file

        Set metadata.

        Attributes have to be of the form documented in the
        `Metadata Reference`_.

        Parameters
        ----------
        kwargs : key-value pairs like field="value", where the values must be
            strings. Does not alter existing fields, unless
            the field appears here - if the value is None, delete the
            field.
        copy_kwargs : dict, optional
            dictionary of additional params to use for the underlying
            s3.copy_object.

        Examples
        --------    
        >>> await mys3file.asetxattr(attribute_1='value1', attribute_2='value2')  # doctest: +SKIP
        # Example for use with copy_args
        >>> await mys3file.asetxattr(copy_kwargs={'ContentType': 'application/pdf'},
        ...     attribute_1='value1')  # doctest: +SKIP

        """
        ...

    async def asetxattr(self, **kwargs):
        try:
            return await self._accessor.asetxattr(self.fspath_, **kwargs)
        except RuntimeError as e:
            if 'Loop is not running' in str(e):
                return await ThreadPool.asyncish(self.setxattr, **kwargs)
            raise e

    # async def asetxattr(self, *args, **kwargs):
    #     return await ThreadPool.asyncish(self.setxattr, *args, **kwargs)
        # return await self._accessor.asetxattr(self.fspath_, *args, **kwargs)

    def invalidate_cache(self):
        return self._accessor.invalidate_cache(self.fspath_)
    
    async def ainvalidate_cache(self):
        return await self._accessor.ainvalidate_cache(self.fspath_)

    def cloze(self, **kwargs):
        if self._fileio:
            self._fileio.commit()
        return self._accessor.invalidate_cache(self.fspath_)

    async def acloze(self, **kwargs):
        return await self._accessor.ainvalidate_cache(self.fspath_)

    def get_checksum(
        self,
        method: str = 'md5',
        chunk_size: int = 1024,
        **kwargs
    ):
        """
        Creates the checksum for the file
        """
        hashmethod = getattr(hashlib, method)
        hasher = hashmethod()
        with self.open('rb') as f:
            for chunk in iter(lambda: f.read(chunk_size), b''):
                hasher.update(chunk)
        checksum = hasher.hexdigest()
        del hasher
        return checksum

    async def aget_checksum(
        self,
        method: str = 'md5',
        chunk_size: int = 1024,
        **kwargs
    ):
        """
        Creates the checksum for the file
        """
        hashmethod = getattr(hashlib, method)
        hasher = hashmethod()
        async with self.aopen('rb') as f:
            if not self.is_fsspec:
                for byte_block in await iter(lambda: f.read(chunk_size), b""):
                    hasher.update(byte_block)
            else:
                byte_block = await f.read(chunk_size)
                while byte_block:
                    hasher.update(byte_block)
                    byte_block = await f.read(chunk_size)
        checksum = hasher.hexdigest()
        del hasher
        return checksum

    @classmethod
    def get_object_size(cls, obj: t.Any) -> ObjectSize:
        """
        Returns the size of the object
        """
        return ObjectSize(obj)


class CloudFileSystemPosixPath(PosixPath, CloudFileSystemPath, PureCloudFileSystemPosixPath):
    __slots__ = ()


class CloudFileSystemWindowsPath(WindowsPath, CloudFileSystemPath, PureCloudFileSystemWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("CloudFileSystemPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("CloudFileSystemPath.async_is_mount() is unsupported on this system")


os.PathLike.register(CloudFileSystemPurePath)
os.PathLike.register(CloudFileSystemPath)
os.PathLike.register(PureCloudFileSystemPosixPath)
os.PathLike.register(CloudFileSystemWindowsPath)
os.PathLike.register(CloudFileSystemPosixPath)
os.PathLike.register(PureCloudFileSystemWindowsPath)

if PYDANTIC_VERSION == 2:
    from pydantic import TypeAdapter
    
def register_pathlike(pathz: t.List[t.Union[PosixPath, CloudFileSystemPath, WindowsPath, CloudFileSystemWindowsPath, CloudFileSystemPosixPath, PureCloudFileSystemWindowsPath, t.Any]]):
    for p in pathz:
        os.PathLike.register(p)
        # if PYDANTIC_VERSION == 2:
        #     TypeAdapter(p)

def register_pydantictype(pathz: t.Union[PosixPath, CloudFileSystemPath, WindowsPath, CloudFileSystemWindowsPath, CloudFileSystemPosixPath, PureCloudFileSystemWindowsPath, t.Any]):
    if PYDANTIC_VERSION != 2: return
    TypeAdapter(pathz)



from __future__ import annotations

"""Utilities for lazily installing and importing optional dependencies."""

import functools
import importlib
import importlib.metadata
import inspect
import os
import pathlib
import shlex
import subprocess
import sys
import typing as t
from types import ModuleType

from lzl.logging import default_logger as logger

__all__ = [
    "PkgInstall",
    "LazyLibType",
    "LazyLib",
    "resolve_missing",
    "resolve_missing_custom",
    "require_missing_wrapper",
]


def get_variable_separator() -> str:
    """Return the PATH separator for the current operating system."""

    return ";" if sys.platform.startswith("win") else ":"


def is_coro_func(obj: t.Any, func_name: str | None = None) -> bool:
    """Return ``True`` when ``obj`` (or ``obj.func_name``) is awaitable."""

    try:
        if inspect.iscoroutinefunction(obj) or inspect.isawaitable(obj):
            return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(
            getattr(obj, func_name)
        ):
            return True
        return bool(hasattr(obj, "__call__") and inspect.iscoroutinefunction(obj.__call__))
    except Exception:  # pragma: no cover - mirrors legacy permissive behaviour
        return False


class PkgInstall:
    """Platform-specific installer command templates."""

    win: str = "choco install [flags]"
    mac: str = "brew [flags] install"
    linux: str = "apt-get -y [flags] install"

    @classmethod
    def get_args(cls, binary: str, flags: list[str] | None = None) -> list[str]:
        """Return a shell-escaped argument list for installing ``binary``."""

        flag_str = " ".join(flags) if flags else ""
        if sys.platform.startswith("win"):
            command = f"{cls.win} {binary}".replace("[flags]", flag_str)
        elif sys.platform.startswith("linux"):
            command = f"{cls.linux} {binary}".replace("[flags]", flag_str)
        else:
            command = f"{cls.mac} {binary}".replace("[flags]", flag_str)
        return shlex.split(command)


_has_uv_available: bool | None = None


def _has_uv() -> bool:
    """Return ``True`` when the ``uv`` CLI is available on PATH."""

    global _has_uv_available
    if _has_uv_available is None:
        _has_uv_available = LazyLib.is_exec_available("uv")
    return _has_uv_available


class LazyLibType(type):
    """Metaclass providing lazy import/install helpers."""

    @classmethod
    def install_binary(cls, binary: str, flags: list[str] | None = None) -> None:
        if cls.get_binary_path(binary):
            return
        args = PkgInstall.get_args(binary, flags)
        subprocess.check_call(args, stdout=subprocess.DEVNULL)

    @classmethod
    def get_requirement(cls, name: str, clean: bool = True) -> str:
        name = name.replace("-", "_")
        return name.split("=")[0].replace(">", "").replace("<", "").strip() if clean else name.strip()

    @classmethod
    def install_library(cls, library: str, upgrade: bool = True) -> None:
        if _has_uv():
            try:
                pip_exec = ["uv", "pip", "install"]
                if "=" not in library or upgrade:
                    pip_exec.append("--upgrade")
                pip_exec.append(library)
                subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)
                return
            except Exception as exc:  # pragma: no cover - fallback path
                logger.warning("Failed to install %s using uv: %s. Falling back to pip.", library, exc)

        pip_exec = [sys.executable, "-m", "pip", "install"]
        if "=" not in library or upgrade:
            pip_exec.append("--upgrade")
        pip_exec.append(library)
        subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)

    @classmethod
    def install_pip_package(
        cls,
        package: str,
        version: str | None = None,
        **options: t.Any,
    ) -> None:
        if _has_uv():
            try:
                if version:
                    if "=" not in version:
                        version = f"=={version}"
                    package = f"{package}{version}"
                pip_exec = ["uv", "pip", "install", package]
                for key, value in options.items():
                    opt = key if key.startswith("--") else f"--{key}"
                    if value and not isinstance(value, bool):
                        pip_exec.append(f"{opt}={value}")
                    else:
                        pip_exec.append(opt)
                subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)
                return
            except Exception as exc:  # pragma: no cover - fallback path
                logger.warning("Failed to install %s using uv: %s. Falling back to pip.", package, exc)

        pip_exec = [sys.executable, "-m", "pip", "install"]
        for key, value in options.items():
            opt = key if key.startswith("--") else f"--{key}"
            if value and not isinstance(value, bool):
                pip_exec.append(f"{opt}={value}")
            else:
                pip_exec.append(opt)
        if version:
            if "=" not in version:
                version = f"=={version}"
            package = f"{package}{version}"
        pip_exec.append(package)
        subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)

    @classmethod
    def is_available(cls, library: str) -> bool:
        try:
            importlib.metadata.version(library)
            return True
        except importlib.metadata.PackageNotFoundError:
            return False

    @classmethod
    def __is_available(cls, library: str) -> bool:
        try:
            import pkg_resources
        except ImportError:
            cls.install_pip_package("setuptools")
            import pkg_resources
        try:
            pkg_resources.get_distribution(library)
            return True
        except pkg_resources.DistributionNotFound:  # pragma: no cover - legacy path
            return False

    @classmethod
    def is_imported(cls, library: str) -> bool:
        return library in sys.modules

    @classmethod
    def _ensure_lib_imported(cls, library: str) -> ModuleType:
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_imported(clean_lib):
            sys.modules[clean_lib] = importlib.import_module(clean_lib)
        return sys.modules[clean_lib]

    @classmethod
    def _ensure_lib_installed(cls, library: str, pip_name: str | None = None, upgrade: bool = False) -> None:
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_available(clean_lib):
            cls.install_library(pip_name or library, upgrade=upgrade)

    @classmethod
    def _ensure_binary_installed(cls, binary: str, flags: list[str] | None = None) -> None:
        cls.install_binary(binary, flags)

    @classmethod
    def import_lib(
        cls,
        library: str,
        pip_name: str | None = None,
        resolve_missing: bool = True,
        require: bool = False,
        upgrade: bool = False,
    ) -> ModuleType:
        clean_lib = cls.get_requirement(library, True)
        if not cls.is_available(clean_lib):
            if require and not resolve_missing:
                raise ImportError(f"Required Lib {library} is not available.")
            if not resolve_missing:
                return None  # type: ignore[return-value]
            cls.install_library(pip_name or library, upgrade=upgrade)
        return cls._ensure_lib_imported(library)

    @classmethod
    def import_module(
        cls,
        name: str,
        library: str | None = None,
        pip_name: str | None = None,
        resolve_missing: bool = True,
        require: bool = False,
        upgrade: bool = False,
    ) -> ModuleType:
        if library:
            cls.import_lib(library, pip_name, resolve_missing, require, upgrade)
            return importlib.import_module(name, package=library)
        return importlib.import_module(name)

    @classmethod
    def import_module_attr(
        cls,
        name: str,
        module_name: str,
        library: str | None = None,
        pip_name: str | None = None,
        resolve_missing: bool = True,
        require: bool = False,
        upgrade: bool = False,
    ) -> t.Any:
        module = cls.import_module(
            name=module_name,
            library=library,
            pip_name=pip_name,
            resolve_missing=resolve_missing,
            require=require,
            upgrade=upgrade,
        )
        return getattr(module, name)

    @classmethod
    def import_cmd(
        cls,
        binary: str,
        resolve_missing: bool = True,
        require: bool = False,
        flags: list[str] | None = None,
    ) -> t.Any:
        if not cls.is_exec_available(binary):
            if require and not resolve_missing:
                raise ImportError(f"Required Executable {binary} is not available.")
            if not resolve_missing:
                return None
            cls.install_binary(binary, flags=flags)
        from lazy.cmd import Cmd  # type: ignore

        return Cmd(binary=binary)

    @classmethod
    def get_binary_path(cls, executable: str) -> str | None:
        if "PATH" not in os.environ:
            return None
        for directory in os.environ["PATH"].split(get_variable_separator()):
            binary = os.path.abspath(os.path.join(directory, executable))
            if os.path.isfile(binary) and os.access(binary, os.X_OK):
                return binary
        return None

    @classmethod
    def is_exec_available(cls, executable: str) -> bool:
        return cls.get_binary_path(executable) is not None

    @staticmethod
    def reload_module(module: ModuleType) -> ModuleType:
        return importlib.reload(module)

    @staticmethod
    def get_cwd(*paths: t.Any, string: bool = True) -> str | pathlib.Path:
        if not paths:
            return pathlib.Path.cwd().as_posix() if string else pathlib.Path.cwd()
        resolved = pathlib.Path.cwd().joinpath(*paths)
        return resolved.as_posix() if string else resolved

    @staticmethod
    def run_cmd(cmd: str, raise_error: bool = True) -> str:
        try:
            out = subprocess.check_output(cmd, shell=True)
            return out.decode("utf8") if isinstance(out, bytes) else out.strip()
        except Exception as exc:
            if not raise_error:
                return ""
            raise exc

    def __getattr__(cls, key: str) -> t.Any:
        if key.startswith("is_avail_bin_") or key.startswith("is_avail_exec_"):
            exec_name = key.split("_", 3)[-1].strip()
            return cls.is_exec_available(exec_name)
        if key.startswith("is_avail_lib_") or key.startswith("is_avail_"):
            lib_name = key.split("is_avail_")[-1].strip()
            return cls.is_available(lib_name)
        if key.startswith("is_imported_"):
            lib_name = key.split("is_imported_")[-1].strip()
            return cls.is_imported(lib_name)
        if key.startswith("cmd_"):
            binary_name = key.split("cmd_")[-1].strip()
            return cls.import_cmd(binary=binary_name)
        return cls.import_lib(key, resolve_missing=False, require=False)

    @classmethod
    def get(
        cls,
        name: str,
        attr_name: str | None = None,
        pip_name: str | None = None,
        resolve_missing: bool = True,
    ) -> ModuleType | t.Any:
        parsed = cls._parse_name(name)
        if attr_name is not None:
            parsed["attr_name"] = attr_name
        if pip_name is not None:
            parsed["pip_name"] = pip_name

        library = t.cast(str, parsed["library"])
        module_name = parsed["module_name"] or library

        if parsed.get("attr_name"):
            return cls.import_module_attr(
                parsed["attr_name"],
                module_name=module_name,
                library=library,
                pip_name=parsed["pip_name"],
                resolve_missing=resolve_missing,
            )
        if parsed.get("module_name"):
            return cls.import_module(
                module_name,
                library=library,
                pip_name=parsed["pip_name"],
                resolve_missing=resolve_missing,
            )
        return cls.import_lib(library, pip_name=parsed["pip_name"], resolve_missing=resolve_missing)

    @classmethod
    def _parse_name(cls, name: str) -> dict[str, str | None]:
        result: dict[str, str | None] = {
            "library": "",
            "pip_name": None,
            "module_name": None,
            "attr_name": None,
        }
        current = name.strip()
        if ":" in current:
            current, attr = current.split(":", 1)
            result["attr_name"] = attr
        if "|" in current:
            pip, current = current.split("|", 1)
            result["pip_name"] = pip
        if "." in current:
            result["module_name"] = current
            result["library"] = current.split(".", 1)[0]
        else:
            result["library"] = current
        return result

    @classmethod
    def __getitem__(cls, name: str) -> ModuleType:
        parsed = cls._parse_name(name)
        library = t.cast(str, parsed["library"])
        module_name = parsed["module_name"] or library
        if parsed.get("attr_name"):
            return cls.import_module_attr(
                parsed["attr_name"],
                module_name=module_name,
                library=library,
                pip_name=parsed["pip_name"],
                resolve_missing=True,
                require=True,
            )
        if parsed.get("module_name"):
            return cls.import_module(
                name=parsed["module_name"],
                library=library,
                pip_name=parsed["pip_name"],
                resolve_missing=True,
                require=True,
            )
        return cls.import_lib(
            library,
            pip_name=parsed["pip_name"],
            resolve_missing=True,
            require=True,
        )


class LazyLib(metaclass=LazyLibType):
    """Concrete helper exposing ``LazyLibType`` utilities via attribute access."""

    pass


def resolve_missing(
    modules: str | list[str],
    packages: str | list[str] | None = None,
    required: bool = True,
) -> None:
    if not isinstance(modules, list):
        modules = [modules]
    if packages is not None and not isinstance(packages, list):
        packages = [packages]
    elif packages is None:
        packages = modules
    kind = "required" if required else "optionally required"
    logger.info("%s are %s. Installing...", ", ".join(modules), kind)
    for module, pkg in zip(modules, packages):
        LazyLib.import_lib(module, pkg)


def resolve_missing_custom(
    modules: str | list[str],
    packages: str | list[str] | None = None,
    required: bool = True,
) -> None:
    if not isinstance(modules, list):
        modules = [modules]
    if packages is not None and not isinstance(packages, list):
        packages = [packages]
    elif packages is None:
        packages = modules

    module_names = [module.split(" ", 1)[0] for module in modules]
    kind = "required" if required else "optionally required"
    logger.info("%s are %s. Installing...", ", ".join(module_names), kind)
    for module, pkg in zip(modules, packages):
        module_name = LazyLib.get_requirement(module, True)
        if LazyLib.is_available(module_name):
            continue
        LazyLib.install_library(pkg)


def require_missing_wrapper(
    resolver: t.Callable[..., None],
    func: t.Callable[..., t.Any],
    **resolver_kwargs: t.Any,
) -> t.Callable[..., t.Any]:
    if is_coro_func(func):

        @functools.wraps(func)
        async def async_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            resolver(**resolver_kwargs)
            return await func(*args, **kwargs)

        return async_wrapper

    @functools.wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        resolver(**resolver_kwargs)
        return func(*args, **kwargs)

    return wrapper


_require_registry: dict[str, bool] = {}



from __future__ import annotations

import yaml
from pathlib import Path
from lzo.types import BaseSettings, eproperty
from lazyops.libs.abcs.types.persistence import TemporaryData
from pydantic import PrivateAttr, model_validator, field_validator, Field
from urllib.parse import urlparse, urlunparse, urljoin, parse_qsl, urlencode, quote
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated, Generator

if TYPE_CHECKING:
    from fastapi import FastAPI
    from .manager import OAuth2Manager

here = Path(__file__).parent

class OAuth2Settings(BaseSettings):
    """
    The Primary OAuth2 Settings
    """

    config_file: Optional[Path] = None
    header: Optional[str] = "Authorization"
    scheme: Optional[str] = "Bearer"

    domains: Optional[List[str]] = None

    # Session Configurations
    session_enabled: Optional[bool] = None

    # User Configurations
    staff_emails: Optional[List[str]] = Field(default_factory = list)
    admin_emails: Optional[List[str]] = Field(default_factory = list)
    sa_emails: Optional[List[str]] = Field(default_factory = list)

    staff_email_domains: Optional[List[str]] = Field(default_factory = list)
    admin_email_domains: Optional[List[str]] = Field(default_factory = list)
    sa_email_domains: Optional[List[str]] = Field(default_factory = list)


    class Config:
        env_prefix = "OAUTH2_"
        arbitrary_types_allowed = True
        extra = 'allow'

    
    @eproperty
    def manager(self) -> 'OAuth2Manager':
        """
        Returns the OAuth2 Manager
        """
        from .manager import OAuth2Manager
        return OAuth2Manager(settings = self)
    
    def get_manager(self) -> 'OAuth2Manager':
        """
        Returns the OAuth2 Manager
        """
        return self.manager

    def load_domains(self, domains: Union[str, List[str], Dict[str, Union[str, List[str]]]]) -> None:
        """
        Loads the domains
        """
        if isinstance(domains, dict):
            domains = self.app_env.select(domains)
        if isinstance(domains, str): domains = [domains]
        self.domains = domains

    def load_user_settings(self, users: Union[str, List[str], Dict[str, Union[str, List[str]]]]) -> None:
        """
        Loads the user settings
        """
        if users.get('admin_emails'):
            self.admin_emails = users['admin_emails']
        if users.get('staff_emails'):
            self.staff_emails = users['staff_emails']
        if users.get('sa_emails'):
            self.sa_emails = users['sa_emails']
        if users.get('admin_email_domains'):
            self.admin_email_domains = users['admin_email_domains']
        if users.get('staff_email_domains'):
            self.staff_email_domains = users['staff_email_domains']
        if users.get('sa_email_domains'):
            self.sa_email_domains = users['sa_email_domains']


    def load_from_config_file(self, config_file: Path) -> None:
        """
        Loads the config from a file
        """
        data: Dict[str, Union[Dict[str, Any], List[Dict[str, Any]], Any]] = yaml.safe_load(config_file.read_text())
        if 'session_enabled' in data:
            self.session_enabled = data['session_enabled']
        if data.get('users'): self.load_user_settings(data['users'])
        if data.get('domains'): self.load_domains(data['domains'])
        if data.get('providers'):
            self.manager.load_providers(data['providers'], domains = self.domains)
        if data.get('routes'): self.manager.routes.update(data['routes'])

    
    @model_validator(mode = 'after')
    def validate_oauth_config(self):
        """
        Validates the config
        """
        if self.config_file and self.config_file.exists():
            self.load_from_config_file(self.config_file)
        return self
    
    @property
    def provider_pattern(self) -> str:
        """
        Returns the provider pattern

        '^(google|kinde|auth0)$'
        """
        if 'provider_pattern' not in self._extra:
            base_pattern = '|'.join(list(self.manager.providers.keys()))
            self._extra['provider_pattern'] = f'^({base_pattern})$'
        return self._extra['provider_pattern']
    
    @property
    def app(self) -> 'FastAPI':
        """
        Returns the app
        """
        return self._extra.get('app')
    
    @app.setter
    def app(self, value: 'FastAPI') -> None:
        """
        Sets the app
        """
        self._extra['app'] = value

    @property
    def temp_data(self) -> 'TemporaryData':
        """
        Returns the temporary data
        """
        if 'temp_data' not in self._extra:
            self._extra['temp_data'] = TemporaryData(
                filepath = here.joinpath('data', 'authz.tmp.json'),
            )
        return self._extra['temp_data']


from __future__ import annotations

"""
Google OAuth2 Config
"""

from ...types.config import BaseOAuth2Config, eproperty
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated
from pydantic import Field
from lzo.utils import create_unique_id
from .token import AccessToken, IDToken, OAuth2Token

_default_scopes = [
    'openid',
    'https://www.googleapis.com/auth/userinfo.email',
    'https://www.googleapis.com/auth/userinfo.profile',
]

class GoogleOAuth2Config(BaseOAuth2Config):
    """
    Google OAuth2 Config
    """
    name: Optional[str] = 'google'
    issuer_url: Optional[str] = None
    jwt_issuer_url: Optional[str] = 'https://accounts.google.com'
    scopes: Optional[List[str]] = Field(default = _default_scopes)
    jwt_url: Optional[str] = 'https://www.googleapis.com/oauth2/v3/certs'

    authorize_disabled: Optional[bool] = True
    user_info_retrieve_disabled: Optional[bool] = True # Disable for now
    
    class Config(BaseOAuth2Config.Config):
        env_prefix = "GOOGLE_"


    @classmethod
    def get_default_token_cls(cls, name: str) -> Union[Type[AccessToken], Type[IDToken], Type[OAuth2Token]]:
        """
        Returns the default token class
        """
        if name == 'access_token':
            return AccessToken
        return IDToken if name == 'id_token' else OAuth2Token


    @eproperty
    def authorization_endpoint(self) -> str:
        """
        Returns the authorization endpoint
        """
        return 'https://accounts.google.com/o/oauth2/v2/auth'
    
    @eproperty
    def token_endpoint(self) -> str:
        """
        Returns the token endpoint
        """
        return 'https://oauth2.googleapis.com/token'
    
    @eproperty
    def token_info_endpoint(self) -> str:
        """
        Returns the token info endpoint
        """
        return 'https://oauth2.googleapis.com/tokeninfo'
    
    @eproperty
    def user_info_endpoint(self) -> str:
        """
        Returns the user info endpoint
        """
        return 'https://openidconnect.googleapis.com/v1/userinfo'

    def get_scopes(
        self,
        scopes: Optional[List[str]] = None,
        sep: Optional[str] = ' ',
    ) -> str:
        """
        Returns the scopes
        """
        if not scopes: return self.scope
        for n, scope in enumerate(scopes):
            if scope == 'openid': continue
            # This deals with `userinfo.email` and `userinfo.profile`
            if not scope.startswith('https://'):
                scopes[n] = f'https://www.googleapis.com/auth/{scope}'
        return sep.join(scopes)

    def get_auth_params(
        self,
        redirect_url: str,
        scope: Optional[str] = None,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the auth params
        """
        scope = scope or self.get_scopes()
        state = state or create_unique_id(alph_only=True)
        return {
            'redirect_uri': redirect_url,
            'client_id': self.client_id,
            'scope': scope,
            'access_type': 'offline',
            'prompt': 'consent',
            'response_type': 'code',
            'state': state,
            'include_granted_scopes': 'true',
        }
    
    def get_fetch_token_params(
        self,
        code: str,
        redirect_url: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the fetch token params
        """
        return {
            'code': code,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri': redirect_url,
            'grant_type': 'authorization_code',
        }
    
    def get_refresh_token_params(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the refresh token params
        """
        return {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
        }





from .methods import (
    ASYNC_METHODS,
    ASYNC_MAPPING_GS,
    ASYNC_MAPPING_S3,
    ASYNC_MAPPING_AZ,
    ASYNC_SYNTAX_MAPPING,
)
from .prefixes import (
    GIT_PREFIXES,
    URI_PREFIXES,
    URI_SCHEMES,
    URI_MAP_ROOT,
    PROVIDER_MAP,
)




from __future__ import annotations

"""Mapping utilities that lazily initialise components on first use."""

import collections.abc
import typing as t

KT = t.TypeVar("KT")
VT = t.TypeVar("VT")
DictValue = t.Union[VT, t.Callable[[], VT], t.Tuple[str, t.Dict[str, t.Any]], str, t.Any]


class ProxyDict(collections.abc.MutableMapping, t.MutableMapping[KT, VT]):
    """Dictionary-like container that defers constructing its values."""

    _dict: t.Dict[KT, DictValue] = {}
    _initialized: t.Dict[KT, bool] = {}
    _excluded_attrs: t.Set[str] = set()
    _pdict_attrs: t.Set[str] = {
        "get_or_init",
        "keys",
        "values",
        "items",
        "get",
        "clear",
        "pop",
        "popitem",
        "setdefault",
        "initialize_objects",
        "update",
        "post_init",
        "pre_init",
        "cls_init",
        "cls_post_init",
        "proxy_schema",
        "module",
        "components",
        "_init_component",
        "_prevalidate_component",
        "_init_default",
    }

    module: str | None = None
    components: t.Optional[t.List[str]] = None
    proxy_schema: t.Optional[t.Dict[str, str]] = {}
    initialize_objects: bool = False

    def __init__(
        self,
        initialize_objects: bool | None = None,
        module: str | None = None,
        components: t.Optional[t.List[str]] = None,
        proxy_schema: t.Optional[t.Dict[str, str]] = None,
        **kwargs: t.Any,
    ) -> None:
        """Configure module metadata and optionally pre-initialise components."""

        self.cls_init(**kwargs)
        if initialize_objects is not None:
            self.initialize_objects = initialize_objects
        if module is not None:
            self.module = module
        if components is not None:
            self.components = components
        if proxy_schema is not None:
            self.proxy_schema = proxy_schema
        self.cls_post_init(**kwargs)
        self.pre_init(**kwargs)
        self.post_init(**kwargs)

    def cls_init(self, **kwargs: t.Any) -> None:
        """Hook executed before instance level configuration is applied."""

    def cls_post_init(self, **kwargs: t.Any) -> None:
        """Hook executed after class-level configuration has run."""

    def pre_init(self, **kwargs: t.Any) -> "ProxyDict[KT, VT]":
        """Normalise component metadata before instance initialisation."""

        if self.components:
            for component in self.components:
                if component not in self.proxy_schema:
                    ref_name = f"{self.module}.{component}" if self.module else component
                    self.proxy_schema[component] = ref_name
        return self

    def post_init(self, **kwargs: t.Any) -> None:
        """Extension hook for subclasses; no-op by default."""

    def __init_subclass__(cls: type["ProxyDict"], **kwargs: t.Any) -> None:
        """Reset class-level caches for each subclass and track custom attrs."""

        super().__init_subclass__(**kwargs)
        cls._dict = {}
        cls._initialized = {}
        for attr in dir(cls):
            if attr in cls._pdict_attrs or attr.startswith("_"):
                continue
            cls._pdict_attrs.add(attr)

    def _init_component(self, name: KT, default: VT | None = None) -> None:
        """Subclasses can override to eagerly populate ``name``."""

    def _prevalidate_component(self, name: KT) -> None:
        """Validate ``name`` before attempting to initialise it."""

    def _init_default(self, name: KT) -> None:
        """Materialise the underlying object for ``name`` based on schema hints."""

        from lzl.load import lazy_import

        value = self._dict[name]
        if isinstance(value, str):
            value = lazy_import(value)
            if self.initialize_objects:
                value = value()
        elif isinstance(value, tuple):
            obj_class, kwargs = value
            if isinstance(obj_class, str):
                obj_class = lazy_import(obj_class)
            for key, val in kwargs.items():
                if callable(val):
                    kwargs[key] = val()
            value = obj_class(**kwargs)
        elif isinstance(value, dict):
            value = type(self)(value)
        elif isinstance(value, type):
            value = value()
        self._dict[name] = value
        self._initialized[name] = True

    def get_or_init(self, name: KT, default: VT | None = None) -> VT:
        """Return ``name`` initialising it when absent."""

        if name not in self._dict:
            self._prevalidate_component(name)
            if name in (self.proxy_schema or {}):
                self._init_component(name, default)
                self._initialized[name] = True
            elif default is not None:
                self._dict[name] = default
            else:
                raise ValueError(f"Default value for {name} is None")
        if name not in self._initialized:
            self._init_default(name)
        return t.cast(VT, self._dict[name])

    def __getattr__(self, name: str) -> VT:
        """Proxy attribute access into the backing dictionary."""

        if (
            (name.startswith("__") and name.endswith("__"))
            or name.startswith("_r")
            or name in self._excluded_attrs
            or name in self._pdict_attrs
        ):
            return super().__getattr__(name)  # type: ignore[misc]
        return self.get_or_init(t.cast(KT, name), None)

    def __setattr__(self, name: str, value: VT) -> None:
        """Store ``value`` in the backing dictionary unless reserved."""

        if (
            (name.startswith("__") and name.endswith("__"))
            or name.startswith("_r")
            or name in self._excluded_attrs
            or name in self._pdict_attrs
        ):
            super().__setattr__(name, value)
            return
        self._dict[t.cast(KT, name)] = value

    def __getitem__(self, name: KT) -> VT:
        return self.get_or_init(name, None)

    def __setitem__(self, name: KT, value: VT) -> None:
        self._dict[name] = value

    def __contains__(self, name: KT) -> bool:  # type: ignore[override]
        return name in self._dict

    def __repr__(self) -> str:
        return repr(self._dict)

    def __str__(self) -> str:
        return str(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> t.Iterator[KT]:
        return iter(self._dict)

    def __delitem__(self, key: KT) -> None:
        del self._dict[key]

    def keys(self) -> t.KeysView[KT]:
        return self._dict.keys()

    def values(self) -> t.ValuesView[VT]:
        return t.cast(t.ValuesView[VT], self._dict.values())

    def items(self) -> t.ItemsView[KT, VT]:
        return t.cast(t.ItemsView[KT, VT], self._dict.items())

    def get(self, name: KT, default: VT | None = None) -> VT:
        return self.get_or_init(name, default)


__all__ = ["ProxyDict", "DictValue"]



from __future__ import annotations

from fastapi import Security
from fastapi.security.api_key import APIKeyHeader
from fastapi.security.http import HTTPBase, HTTPAuthorizationCredentials
from typing import Optional, Annotated


AuthorizationScheme = HTTPBase(
    scheme = "bearer",
    scheme_name = "Authorization",
    description = "Bearer Token Authentication",
    auto_error = False,
)

APIKeyScheme = APIKeyHeader(
    name = "x-api-key",
    scheme_name = "APIKey",
    description = "API Key Authentication",
    auto_error = False,
)
APIKey = Annotated[Optional[str], Security(APIKeyScheme)]
Authorization = Annotated[Optional[HTTPAuthorizationCredentials], Security(AuthorizationScheme)] 


import sys
from pathlib import Path
from setuptools import setup

root = Path(__file__).parent
version_text = root.joinpath('src', 'version.py').read_text()
version = version_text.split('VERSION = ', 1)[-1].strip().replace('-', '').replace("'", '')

if sys.version_info.minor < 8:
    pass  # Dependencies are handled in pyproject.toml with markers

setup(version=version)


from __future__ import annotations

import datetime
from pydantic import BaseModel, model_validator, Field
from typing import Optional, Dict, Any, Union, List

class SlackContext(BaseModel):
    """
    Holds the Slack Context
    """

    channels: Dict[str, str] = Field(default_factory = dict)
    users: Dict[str, str] = Field(default_factory = dict)
    user_info: Dict[str, Dict[str, Any]] = Field(default_factory = dict)
    user_lookup: Dict[str, str] = Field(default_factory = dict)
    uids: Dict[str, str] = Field(default_factory = dict)

    username_mapping: Optional[Dict[str, Union[str, List[str]]]] = Field(default_factory = dict)
    initialized: Optional[bool] = False
    
    def lookup_id(self, name: str) -> Optional[str]:
        """
        Lookup ID
        """
        return self.uids.get(name, self.users.get(name, self.channels.get(name)))
    

def create_slack_timestamp() -> int:
    """
    Creates a Slack Timestamp
    """
    return int(datetime.datetime.now().timestamp())


class SlackPayload(BaseModel):
    token: Optional[str] = None
    team_id: Optional[str] = None
    team_domain: Optional[str] = None
    enterprise_id: Optional[str] = None
    enterprise_name: Optional[str] = None
    channel_id: Optional[str] = None
    channel_name: Optional[str] = None
    user_id: Optional[str] = None
    user_name: Optional[str] = None
    command: Optional[str] = None
    text: Optional[str] = None
    response_url: Optional[str] = None
    trigger_id: Optional[str] = None
    api_app_id: Optional[str] = None

    ccommand: Optional[str] = None # A mutable copy of the command
    ctext: Optional[str] = None # A mutable copy of the text

    timestamp: Optional[int] = Field(default_factory = create_slack_timestamp)
    
    @model_validator(mode = 'after')
    def set_mutable_context(self):
        """
        Sets the ctext
        """
        if self.text is not None: self.ctext = self.text
        if self.command is not None: self.ccommand = self.command
        return self
    
    def from_repeat(self, payload: SlackPayload) -> 'SlackPayload':
        """
        Sets the existing payload from a previous payload
        """
        self.text = payload.text
        self.command = payload.command
        self.ctext = payload.ctext
        self.ccommand = payload.ccommand
        return self



from __future__ import annotations

import typing as t
from functools import lru_cache

if t.TYPE_CHECKING:
    from pydantic.types import ByteSize

__all__ = ['parse_memory_metric', 'parse_memory_metric_to_bs', 'safe_float']


def parse_memory_metric(data: t.Union[int, float, str]) -> float:
    """Convert Kubernetes-style memory strings (``Mi``, ``Gi``) into bytes."""

    if isinstance(data, (int, float)): return data
    if data.endswith('B'): 
        data = data[:-1]
    cast_type = float if '.' in data else int
    if data.endswith('Ki') or data.endswith('KiB'):
        return cast_type(data.split('Ki')[0].strip()) * 1024
    if data.endswith('Mi') or data.endswith('MiB'):
        return cast_type(data.split('Mi')[0].strip()) * (1024 * 1024)
    if data.endswith('Gi') or data.endswith('GiB'):
        return cast_type(data.split('Gi')[0].strip()) * (1024 * 1024 * 1024)
    if data.endswith('Ti') or data.endswith('TiB'):
        return cast_type(data.split('Ti')[0].strip()) * (1024 * 1024 * 1024 * 1024)
    if data.endswith('Pi') or data.endswith('PiB'):
        return cast_type(data.split('Pi')[0].strip()) * (1024 * 1024 * 1024 * 1024 * 1024)
    if data.endswith('Ei') or data.endswith('EiB'):
        return cast_type(data.split('Ei')[0].strip()) * (1024 * 1024 * 1024 * 1024 * 1024 * 1024)
    return cast_type(data.strip())


def parse_memory_metric_to_bs(data: t.Union[int, float, str]) -> 'ByteSize':
    """Return a :class:`ByteSize` instance from byte or string input."""

    from pydantic.types import ByteSize
    if isinstance(data, (int, float)): 
        return ByteSize(data)
    data = parse_memory_metric(f'{data} MiB')
    return ByteSize(data)


def safe_float(item: str) -> float:
    """Return ``float(item)`` or ``nan`` when conversion fails."""

    try:
        number = float(item)
    except ValueError:
        number = float('nan')
    return number
